# import os
# import sys
from dataclasses import dataclass
from typing import Self #, Optional, Union, Sequence
# import numpy as np
from astropy import coordinates as coord





@dataclass
class Source:
    name: str
    ivsname: str
    coordinates: coord.SkyCoord
    fluxes: dict[str, list[float]]
    category: str
    url: str




class CalibratorCatalog:
    """Full RFC catalog with all sources.
    """
    @property
    def sources(self) -> dict:
        """Returns all sources in the catalog.
        """
        raise NotImplementedError


    def unresolved(self) -> Self:
        """Returns all sources that remain unresolved at long baselines.
        """
        raise NotImplementedError


    def brighter_than(self, flux: float, band: str) -> Self:
        """Returns all sources that are brighter than the given amount on the
        longest baselines.

        Input
        - flux : float
            Flux density on the longest baselines in Jansky units.
        - band : str
            Band at which the flux is compared. It needs to be 'S', 'C', 'X', 'U', 'K'
        """
        raise NotImplementedError


    def get(self, source_name: str) -> Source:
        """Returns the source associated to the given name.
        """
        raise NotImplementedError


    def import_rfc(self):
        """Imports the RFC catalog of sources
        """
        raise NotImplementedError


    def import_vsop(self):
        """Imports the VSOP catalog of sources
        """
        raise NotImplementedError


    def import_icrf(self):
        """Imports the ICRF catalog of sources
        """
        raise NotImplementedError







import numpy as np
from datetime import datetime as dt
import enum
from typing import Optional, Union
from astropy import units as u
from fpdf import FPDF
from dash import dcc
from dash import html
import dash_bootstrap_components as dbc
import plotly.express as px
from vlbiplanobs import freqsetups as fs
from vlbiplanobs import stations
from vlbiplanobs import observation




class SourceEpoch(enum.Enum):
    """Enum-type class with the three possible types of observations to set,
    in terms of how to specify the source and epoch to observe:
    - UNSPECIFIED: if neither source nor epoch have been specified.
    - ONLY_SOURCE: only source is specified, not the epoch.
    - SOURCE_AND_EPOCH: both the epoch and the duration are specified.
    """
    UNSPECIFIED = 0
    ONLY_SOURCE = 1
    SOURCE_AND_EPOCH = 2



def tooltip(message: str, idname: str, trigger: str = '?', placement: str = 'right',
            trigger_is_sup: bool = True,  **kwargs) -> list:
    """Defines a tooltip (popover) that will be shown in the rendered page.
    It will place a <sup>`trigger`</sup> in the page within a span tag.
    Returns the list with all html elements.
    """
    if trigger_is_sup:
        return [html.Span(children=html.Sup(trigger, className='popover-link'), id=idname),
            dbc.Tooltip(message, target=idname, placement=placement, **kwargs)]
    else:
        return [html.Span(children=trigger, className='popover-link', id=idname),
            dbc.Tooltip(message, target=idname, placement=placement, **kwargs)]


def tooltip_card(a_card: dbc.Card, idname: str, trigger: str, placement: str = 'right', **kwargs) -> list:
    """Defines a tooltip (popover) that shows a card.
    """
    return [html.Span(children=trigger, className='popover-link', id=idname),
            dbc.Tooltip(a_card, target=idname, placement=placement,
                innerClassName='tooltip-card-inner', hide_arrow=True,
                **kwargs)]


def create_accordion_card(title: str, text: str, id: str, is_open: bool = True) -> dbc.Card:
    """Given a title (header) and a text (which can be either text, a dcc/html object),
    it will return a dbc.Card object which is one input for an accordion.
    """
    card_header = dbc.CardHeader(html.H2(dbc.Button(title, color='link',
                    id=f"group-{id}-toggle", className='')), className='accordion-header')
    card_body = dbc.Collapse(dbc.CardBody(text), id=f"collapse-{id}", is_open=is_open,
                    className='accordion-collapse')
    return dbc.Card([card_header, card_body], className='accordion-card')


def create_sensitivity_card(title: str, message: Union[str, list]) -> list:
    """Defines one of the cards that are shown in the Sensitivity tab. Each tab
    shows a title `title` and a message `message`.
    If message is a list (of strings), it is assumed as different paragraphs.
    It returns the HTML code of the card.
    """
    ps = []
    if type(message) is list:
        for a_msg in message:
            ps.append(html.P(className='card-text', children=a_msg))
    else:
        ps = [html.P(className='card-text', children=message)]

    # return [html.Div(dbc.Card(
    return [dbc.Card(
            dbc.CardBody([
                html.H5(className='card-title', children=title),
                *ps
                ]), className='card-summary shadow-1-strong')
            ]
    # return [html.Div(className='card-summary m-3', children=[
    #         html.Div(className='card-body', children=[
    #             html.H5(className='card-title', children=title)] + ps)])
    #     ]




#################################################################################
# It is all about cards in this section

def antenna_card(app, station: stations.Station) -> dbc.Card:
    """Generates a card showing the basic information for the given station
    """
    s = lambda st : st[::-1].replace(' ,',' dna ',1)[::-1]
    card = dbc.Card([
        dbc.CardImg(src=app.get_asset_url(f"ant-{station.name.replace(' ','_').lower()}.jpg"),
                    top=True, className='card-img'),
        dbc.CardBody([
            html.H4(station.name, className='card-title'),
            html.H6(station.fullname if station.fullname != station.name else '',
                    className='card-title2'),
            html.H6([html.Span(station.country),
                     html.Span(station.diameter, style={'float': 'right'}),
                    ], className='card-subtitle'),
            # html.P(f"&#127462; Participates in {station.all_networks}.\n"
            dcc.Markdown([f"Listed for the {s(station.all_networks)}.\n" if \
                          station.all_networks != '' else '', "Can observe at "
                         f"{', '.join([i.replace('cm', '') for i in station.bands])} cm."],
                         className='card-text')
            ])
        ], className='card-antenna')
    return card



def antenna_cards(app, stations: list) -> dbc.Row:
    cards = dbc.Row([antenna_card(app, s) for s in stations],
            className='row justify-content-center')
    return cards



def network_card(app, network_acr: str, network_name: str, body,
                 network_img: Optional[str] = None) -> dbc.Card:
    """Generates a card showing the basic information from a network.
    """
    card = dbc.Card([
        dbc.CardImg(src=app.get_asset_url(network_img if network_img is not None else \
                                          f"network-{network_acr.lower().replace(' ', '_')}.png"),
                    top=True, className='card-img'),
        dbc.CardBody([
            html.H4(network_acr, className='card-title'),
            html.H6(network_name, className='card-title2'),
            *[dbc.Checklist(id='initial-e-EVN', className='checkbox', persistence=True,
                    options=[{'label': 'e-EVN (real time) mode', 'value': False}]) \
              if network_acr == 'EVN' else html.Span()],
            ]),
        dbc.CardFooter(
            dbc.Checklist(id=f'network-{network_acr.lower()}', className='checkbox', persistence=True,
                          options=[{'label': ' Select network',
                                    'value': True}], value=[]))
        ], className="card-network col-sm-3 m-4 shadow-1-strong")
    return card



def summary_card_worldmap(app, obs: observation.Observation) -> html.Div:
    """Generates a world map with the participating stations
    """
    ants_up = obs.is_visible()
    ant_no_obs = []
    for an_ant in ants_up:
        if len(ants_up[an_ant][0]) == 0:
            ant_no_obs.append(an_ant)

    return worldmap_plot([obs.stations[a] for a in obs.stations.codenames \
                if a not in ant_no_obs])



def summary_card_antennas(app, obs: observation.Observation) -> list:
    """Generates the summary card with the information about which
    antennas can observe the given source, and the longest/shortest baselines.
    """
    ants_up = obs.is_visible()
    ant_no_obs = []
    for an_ant in ants_up:
        if len(ants_up[an_ant][0]) == 0:
            ant_no_obs.append(an_ant)

    ant_text = ', '.join([ant for ant in obs.stations.codenames if ant not in ant_no_obs]) + '.'
    ant_text = []
    for ant in obs.stations.codenames:
        if ant not in ant_no_obs:
            ant_text += tooltip_card(antenna_card(app, obs.stations[ant]),
                            idname=f"basel-{ant}", trigger=ant, placement='top')
            ant_text += [html.Span(", ")]

    # Remove the trailing ,
    if len(ant_text) > 0:
        ant_text[-1] = html.Span(".")
    else:
        ant_text = ["None."]
    temp_msg = []
    temp_msg += [[f"{len(ants_up)-len(ant_no_obs)} participating antennas: ", *ant_text]]
    if len(ant_no_obs) > 0:
        ant_text = []
        for ant in ant_no_obs:
            ant_text += tooltip_card(antenna_card(app, obs.stations[ant]),
                            idname=f"basel-{ant}", trigger=ant, placement='top')
            ant_text += [html.Span(", ")]

        temp_msg += [html.P(className='text-danger', children=["Note that ", *ant_text[:-1],
                            " cannot observe the source during the planned observation."])]

    longest_bl = obs.longest_baseline()
    ant_l1, ant_l2 = longest_bl[0].split('-')
    # Using dummy units to allow the conversion
    longest_bl_lambda = longest_bl[1]/obs.wavelength
    longest_bl_lambda = optimal_units(longest_bl_lambda*u.m, [u.Gm, u.Mm, u.km])
    temp_msg += [[*baseline_img(app, is_long=True),
                *tooltip_card(antenna_card(app, obs.stations[ant_l1]), idname='basel-l1',
                             trigger=ant_l1, placement='top'),
                html.Span("-"),
                *tooltip_card(antenna_card(app, obs.stations[ant_l2]), idname='basel-l2',
                             trigger=ant_l2, placement='top'),
                f" is the longest (projected) baseline with {optimal_units(longest_bl[1], [u.km, u.m]):.5n} ({longest_bl_lambda.value:.3n} {longest_bl_lambda.unit.name[0]}\u03BB)."]]

    shortest_bl = obs.shortest_baseline()
    ant_s1, ant_s2 = shortest_bl[0].split('-')
    # Using dummy units to allow the conversion
    shortest_bl_lambda = shortest_bl[1]/obs.wavelength
    shortest_bl_lambda = optimal_units(shortest_bl_lambda*u.m, [u.Gm, u.Mm, u.km])
    temp_msg += [[*baseline_img(app, is_long=False),
                *tooltip_card(antenna_card(app, obs.stations[ant_s1]), idname='basel-s1',
                             trigger=ant_s1, placement='top'),
                html.Span("-"),
                *tooltip_card(antenna_card(app, obs.stations[ant_s2]), idname='basel-s2',
                             trigger=ant_s2, placement='top'),
                 f" is the shortest one with {optimal_units(shortest_bl[1], [u.km, u.m]):.5n} ({shortest_bl_lambda.value:.3n} {shortest_bl_lambda.unit.name[0]}\u03BB)."]]
    # )]
    return create_sensitivity_card('Antennas', temp_msg)


def summary_card_beam(app, obs):
    """Creates a summary card showing the expected synthesized beam.
    """
    synthbeam = obs.synthesized_beam()
    synthbeam_units = optimal_units(synthbeam['bmaj'], [u.arcsec, u.mas, u.uas]).unit
    temp_msg = [html.Div(className='row', style={'height': '1rem'}),
                html.Div(className='row justify-content-center',
              children=[ellipse(bmaj="5rem",
                        bmin=f"{5*synthbeam['bmin'].to(u.mas)/synthbeam['bmaj'].to(u.mas)}rem",
                        pa=f"{-synthbeam['pa'].to(u.deg).value+90}deg")])]
    temp_msg += [html.Br(), html.P([f"The expected synthesized beam will be approx. {synthbeam['bmaj'].to(synthbeam_units).value:.3n} x {synthbeam['bmin'].to(synthbeam_units):.3n}", html.Sup("2"), \
            f", PA = {synthbeam['pa']:.3n}."])]
    temp_msg += [html.P("Note that the synthesized beam can significantly change depending "
                        "on the weighting used during imaging (natural weighting assumed here).")]
    return create_sensitivity_card('Resolution', temp_msg)


def summary_card_times(app, obs):
    """Creates a summary card showing the observing times, and the resulting data size.
    """
    prtobstimes = obs.print_obs_times()
    if '\n' in prtobstimes:
        tmp = [html.Span(t) for t in obs.print_obs_times().split('\n')]
        for i in range(len(tmp)-1):
            tmp.insert(2*i+1, html.Br())
            temp_msg = [tmp]
    else:
        temp_msg = [f"{obs.print_obs_times()}."]

    temp_msg += [f"The observation lasts for {optimal_units(obs.duration, [u.h, u.min, u.s, u.ms]):.3n}, of which {optimal_units(obs.ontarget_time, [u.h, u.min, u.s, u.ms]):.3n} are on target."]
    n_files = int(np.ceil(obs.datasize()/(2.0*u.GB)))
    if n_files < 10:
        img_fits = [html.Img(src=app.get_asset_url("icon-file.svg"), height='35rem',
                    style={'display': 'inline'}) for i in range(n_files)]
    else:
        img_fits = [html.Img(src=app.get_asset_url("icon-file.svg"), height='35rem',
                    style={'display': 'inline'}) for i in range(10)]
        img_fits += ["+"]

    temp_msg += [[*img_fits, html.Br(),
        f"With a time integration of {optimal_units(obs.inttime, [u.s,u.ms,u.us]):.2n} the "
        f"expected FITS file size is "
        f"{optimal_units(obs.datasize(), [u.TB, u.GB, u.MB, u.kB]):.3n}. "]]
        # f"(divided in {n_files} 2-GB files)."]]
    return create_sensitivity_card('Observing Time', temp_msg)


def summary_card_frequency(app, obs):
    pol_dict = {1: 'single', 2: 'dual', 4: 'full'}
    bw = optimal_units(obs.bandwidth, [u.GHz, u.MHz, u.kHz])
    vel = optimal_units((2.9979e5*u.km/u.s)*(obs.bandwidth/(2*obs.frequency)).decompose(), [u.km/u.s, u.m/u.s])
    bwwl = optimal_units((30*u.cm/(obs.bandwidth.to(u.GHz).value)), [u.m, u.cm, u.mm])
    temp_msg = [html.Div(className='row justify-content-center',
                children=[html.Div(className='bandpass', style={'height': '4rem',
                'width': f"{90/obs.subbands}%"})]*obs.subbands)]
    temp_msg += [html.Div(className='row justify-content-center',
                 children=[html.Table(className='baseline-table',
                           style={'width': '100%', 'font-size': '0.8rem', 'margin-top':'-1rem'},
                           children=[html.Tr([
                                html.Td(f"-{bw/2:.3n}", style={'text-align': 'left'}),
                                html.Td(f"{optimal_units(obs.frequency, [u.GHz, u.MHz]):.4n}",
                                        style={'text-align': 'center'}),
                                html.Td(f"+{bw/2:.3n}", style={'text-align': 'right'})
                            ]),
                            html.Tr([
                                html.Td(f"+{vel:.5n}", style={'text-align': 'left'}),
                                html.Td("0", style={'text-align': 'center'}),
                                html.Td(f"-{vel:.5n}", style={'text-align': 'right'})
                            ])]
                )])]
    temp_msg += [f"The central frequency is {optimal_units(obs.frequency, [u.GHz, u.MHz]):.3n} ({optimal_units(obs.wavelength, [u.m, u.cm, u.mm]):.2n})."]
    if obs.subbands > 1:
        temp_msg += [f"The total bandwidth of {optimal_units(obs.bandwidth, [u.GHz, u.MHz, u.kHz]):.3n} will be divided into {obs.subbands} subbands of {optimal_units(obs.bandwidth/obs.subbands, [u.GHz, u.MHz, u.kHz]):.3n} each, with {obs.channels} channels ({optimal_units(obs.bandwidth/(obs.subbands*obs.channels), [u.GHz, u.MHz, u.kHz, u.Hz]):.3n}, or {optimal_units(2*vel/(obs.subbands*obs.channels), [u.km/u.s, u.m/u.s]):.3n}, wide)."]
    else:
        temp_msg += [f"The total bandwidth of {optimal_units(obs.bandwidth, [u.GHz, u.MHz, u.kHz]):.3n} will be recorded in {obs.subbands} subband, with {obs.channels} channels ({optimal_units(obs.bandwidth/(obs.subbands*obs.channels), [u.GHz, u.MHz, u.kHz, u.Hz]):.3n}, or {optimal_units(2*vel/(obs.subbands*obs.channels), [u.km/u.s, u.m/u.s]):.3n}, wide)."]
    temp_msg += [f"Recording {pol_dict[obs.polarizations]} circular polarization."]
    return create_sensitivity_card('Frequency Setup', temp_msg)



def summary_card_fov(app, obs):
    """Creates a summary card showing the expected FoV limitations.
    """
    shortest_bl = obs.shortest_baseline()[1]
    largest_ang_scales = ((2.063e8*u.mas)*(obs.wavelength/shortest_bl)).to(u.mas)
    pb_scale = ((2.063e8*u.mas)*(obs.wavelength/(100*u.m))).to(u.arcsec)
    bw_smearing = obs.bandwidth_smearing()
    tm_smearing = obs.time_smearing()
    smearing_ratio = bw_smearing/tm_smearing
    smearing_ratio = smearing_ratio if smearing_ratio <= 1.0 else 1/smearing_ratio
    temp_msg = [html.Div(className='row', style={'height': '1rem'}),
            html.Div(className='row justify-content-center align-self-center',
                    children=[ellipse(bmaj="5rem", bmin="5rem", pa="0deg"),
                      ellipse(bmaj=f"2rem", bmin=f"2rem", pa="0deg", color="#F0959B",
                              z_index=3, position='absolute', className='align-self-center'),
                      ellipse(bmaj=f"{2*smearing_ratio}rem",
                              bmin=f"{2*smearing_ratio}rem",
                              pa="0deg", color="white", z_index=4, position='absolute',
                              className='align-self-center')])]
    temp_msg += [f"The Field of View would be limited by time smearing to a radius of "
                 f"{optimal_units(tm_smearing, [u.arcmin, u.arcsec]):.3n} and by frequency smearing to "
                 f"{optimal_units(bw_smearing, [u.arcmin, u.arcsec]):.3n} (considering a 10% loss), "
                 "if no further time/frequency averaging is performed."]
    temp_msg += [f"Considering the shortest baseline in the array, "
                 "you will resolve out emission on angular scales larger than "
                 f"{optimal_units(largest_ang_scales, [u.arcmin, u.arcsec, u.mas]):.3n}."]
    return create_sensitivity_card('FoV limitations', temp_msg)



def summary_card_rms(app, obs):
    """Creates a summary card showing the reached sensitivity.
    """
    rms = optimal_units(obs.thermal_noise(), [u.MJy, u.kJy, u.Jy, u.mJy, u.uJy])
    rms_time = optimal_units(obs.thermal_noise()/np.sqrt(obs.inttime/obs.duration),
                             [u.MJy, u.kJy, u.Jy, u.mJy, u.uJy])
    rms_channel = optimal_units(rms*np.sqrt(obs.subbands*obs.channels),
                                [u.MJy, u.kJy, u.Jy, u.mJy, u.uJy])
    temp_msg = [html.Div(className='row', style={'height': '0.7rem'}),
                html.Div(className='row justify-content-center',
                children=html.Img(src=app.get_asset_url("waves.png"), width='100%',
                                  height='75rem', style={'display': 'inline'}))]
    temp_msg += [html.P(f"The expected rms thermal noise for your target is {rms:.3n}/beam using natural weighting during imaging. Note that ~20% higher values may be expected for RFI-contaminated bands.")]
    temp_msg += [html.P(f"The achieved sensitivity implies a rms of {rms_channel:.3n}/beam per spectral "
                        f"channel, or approx. {rms_time:.3n}/beam per time integration "
                        f"({optimal_units(obs.inttime, [u.s,u.ms,u.us]):.3n}).")]
    return create_sensitivity_card('Sensitivity', temp_msg)



def summary_printable(app, obs):
    """Returns a html.Div object with a HTML format.
    """
    rms = optimal_units(obs.thermal_noise(), [u.MJy, u.kJy, u.Jy, u.mJy, u.uJy])
    rms_channel = optimal_units(rms*np.sqrt(obs.subbands*obs.channels),
                        [u.MJy, u.kJy, u.Jy, u.mJy, u.uJy])
    bw_smearing = obs.bandwidth_smearing()
    tm_smearing = obs.time_smearing()
    pol_dict = {1: 'single', 2: 'dual', 4: 'full'}
    synthbeam = obs.synthesized_beam()
    synthbeam_units = optimal_units(synthbeam['bmaj'], [u.arcsec, u.mas, u.uas]).unit
    ants_up = obs.is_visible()
    ant_no_obs = []
    for an_ant in ants_up:
        if len(ants_up[an_ant][0]) == 0:
            ant_no_obs.append(an_ant)

    ants = [ant for ant in obs.stations.codenames if ant not in ant_no_obs]
    if obs.target is not None:
        target = f"Target source: {obs.target.coord.to_string('hmsdms')}{' ('+obs.target.name+').' if obs.target.name != 'Source' else '.'}"
    else:
        target = "Target source: unspecified."

    epoch = f"{obs.print_obs_times()}" if obs._fixed_time else \
        f"Observing epoch: unspecified.\nTotal observing time: {optimal_units(obs.duration, [u.h, u.min, u.s]):.3n}"
    # Create the PDF and returns it
    pdf = FPDF(orientation='P', unit='mm', format='A4')
    pdf.compress = False
    pdf_w = 210
    pdf_h = 297
    # self.image(sctplt,  link='', type='', w=1586/80, h=1920/80)
    pdf.add_page()
    pdf.ln(1)
    pdf.set_xy(0.0, 0.0)
    pdf.set_font('Arial', 'B', 16)
    # self.set_text_color(220, 50, 50)
    pdf.cell(w=210.0, h=40.0, align='C', txt="EVN Observation Planner - Summary Report", border=0)
    pdf.set_xy(10.0, 40.0)
    pdf.set_font('Arial', 'B', 12)
    pdf.multi_cell(0, 0, "Schedule")
    pdf.set_xy(10.0, 45.0)
    pdf.set_font('Arial', '', 10)
    pdf.multi_cell(0, 6, f"{epoch}.\n" \
       f"{optimal_units(obs.ontarget_time, [u.h, u.min, u.s, u.ms]):.3n} are on target.\n" \
       f"{target}\n" \
       f"Output FITS file size: {optimal_units(obs.datasize(), [u.TB, u.GB, u.MB, u.kB]):.3n}.\n\n\n")
    pdf.set_xy(10.0, 85.0)
    pdf.set_font('Arial', 'B', 12)
    pdf.multi_cell(0, 0, "Frequency Setup")
    pdf.set_xy(10.0, 90.0)
    pdf.set_font('Arial', '', 10)
    pdf.multi_cell(0, 6, f"Central frequency: {optimal_units(obs.frequency, [u.GHz, u.MHz]):.3n} " \
       f"({optimal_units(obs.wavelength, [u.m, u.cm, u.mm]):.2n}).\n" \
       f"{obs.subbands} subbands of {optimal_units(obs.bandwidth/obs.subbands, [u.GHz, u.MHz, u.kHz]):.3n} each.\n" \
       f"Channels per subband: {obs.channels}.\n" \
       f"Polarization: {pol_dict[obs.polarizations]}.\n" \
       f"time integration: {optimal_units(obs.inttime, [u.s,u.ms,u.us]):.2n}.\n\n\n")
    pdf.set_xy(10.0, 135.0)
    pdf.set_font('Arial', 'B', 12)
    pdf.multi_cell(0, 0, "VLBI Network")
    pdf.set_xy(10.0, 140.0)
    pdf.set_font('Arial', '', 10)
    pdf.multi_cell(0, 6, f"Participating antennas: {', '.join(ants)}.\n" \
       f"The expected synthesized beam will be approx. {synthbeam['bmaj'].to(synthbeam_units).value:.3n} x {synthbeam['bmin'].to(synthbeam_units):.3n}, PA = {synthbeam['pa']:.3n}.\n" \
       f"Expected rms thermal noise level: {rms:.3n}/beam.\n" \
       f"Per spectral channel: {rms_channel:.3n}/beam.\n" \
       f"Time smearing (10% loss): {optimal_units(tm_smearing, [u.arcmin, u.arcsec]):.3n}.\n" \
       f"Frequency smearing (10% loss): {optimal_units(bw_smearing, [u.arcmin, u.arcsec]):.3n}.\n")

    return pdf
    # Single string version
    return "EVN Observation Planner - Summary Report\n\n\n" \
           "- Schedule\n\n" \
           f"{obs.print_obs_times()}.\n" \
           f"{optimal_units(obs.ontarget_time, [u.h, u.min, u.s, u.ms]):.3n} are on target.\n" \
           f"{target}\n" \
           f"FITS file size: {optimal_units(obs.datasize(), [u.TB, u.GB, u.MB, u.kB]):.3n}.\n\n\n" \
           "- Frequency Setup\n\n" \
           f"Central frequency: {optimal_units(obs.frequency, [u.GHz, u.MHz]):.3n} " \
           f"({optimal_units(obs.wavelength, [u.m, u.cm, u.mm]):.2n}).\n" \
           f"{obs.subbands} subbands of {optimal_units(obs.bandwidth/obs.subbands, [u.GHz, u.MHz, u.kHz]):.3n} each.\n" \
           f"Channels per subband: {obs.channels}.\n" \
           f"Polarization: {pol_dict[obs.polarizations]}.\n" \
           f"time integration: {optimal_units(obs.inttime, [u.s,u.ms,u.us]):.2n}.\n\n\n" \
           "- VLBI Network\n\n" \
           f"Participating antennas: {ants}.\n" \
           f"The expected synthesized beam will be approx. {synthbeam['bmaj'].to(synthbeam_units).value:.3n} " \
           f"x {synthbeam['bmin'].to(synthbeam_units):.3n}, PA = {synthbeam['pa']:.3n}.\n" \
           f"Expected rms thermal noise level: {rms:.3n}/beam.\n" \
           f"Per spectral channel: {rms_channel:.3n}/beam.\n" \
           f"Time smearing (10% loss): {optimal_units(tm_smearing, [u.arcmin, u.arcsec]):.3n}\n" \
           f"Frequency smearing (10% loss): {optimal_units(bw_smearing, [u.arcmin, u.arcsec]):.3n}\n" \
           f"Per spectral channel: {rms_channel:.3n}/beam.\n"
    # Dash version
    return html.Div([
            html.H3("EVN Observation Planner - Report"),
            html.H5("Schedule"),
            html.P(f"{obs.print_obs_times()}."),
            html.P(f"{optimal_units(obs.ontarget_time, [u.h, u.min, u.s, u.ms]):.3n} are on target."),
            html.P(f"Target source: {obs.target.coord.to_string('hmsdms')}"
                   f"{' ('+obs.target.name+').' if obs.target.name != 'Source' else '.'}"
                   if obs.target is not None else 'Target source: Unspecified.'),
            html.P(f"FITS file size: {optimal_units(obs.datasize(), [u.TB, u.GB, u.MB, u.kB]):.3n}."),

            html.H5("Frequency Setup"),
            html.P(f"Central frequency: {optimal_units(obs.frequency, [u.GHz, u.MHz]):.3n} "
                   f"({optimal_units(obs.wavelength, [u.m, u.cm, u.mm]):.2n})."),
            html.P(f"{obs.subbands} subbands of "
                   f"{optimal_units(obs.bandwidth/obs.subbands, [u.GHz, u.MHz, u.kHz]):.3n} each."),
            html.P(f"Channels per subband: {obs.channels}."),
            html.P(f"Polarization: {pol_dict[obs.polarizations]}."),
            html.P(f"time integration: {optimal_units(obs.inttime, [u.s,u.ms,u.us]):.2n}."),
            html.P(f""),

            html.H5("VLBI Network"),
            html.P(f"Participating antennas: {ants}."),
            html.P("The expected synthesized beam will be approx. "
                   f"{synthbeam['bmaj'].to(synthbeam_units).value:.3n} x "
                   f"{synthbeam['bmin'].to(synthbeam_units):.3n}, PA = {synthbeam['pa']:.3n}."),
            html.P(f"Expected rms thermal noise level: {rms:.3n}/beam."),
            html.P(f"Per spectral channel: {rms_channel:.3n}/beam."),
            html.P(f"Time smearing (10% loss): {optimal_units(tm_smearing, [u.arcmin, u.arcsec]):.3n}"),
            html.P(f"Frequency smearing (10% loss): {optimal_units(bw_smearing, [u.arcmin, u.arcsec]):.3n}"),
            html.P(f"Per spectral channel: {rms_channel:.3n}/beam."),
            # TODO: Add figure elevations per antenna
        ])


#################################################################################
# Some small graphical elements


def ellipse(bmaj, bmin, pa, color='#a01d26', z_index=1, position='relative', margin_top='', className=''):
    """Returns a html.Div element that draws an ellipse with a semimajor axis bmaj,
    semiminor axis bmin, and position angle (as defined in radio astronomy) pa.
    bmaj,bmin, pa must be strings recognized by HTML/CSS.
    """
    return html.Div(children=[], style={'width': bmaj, 'height': bmin,
                    'border-radius': '50%', 'background': color, 'position': position,
                    'transform': f"rotate({pa})", 'z-index': z_index,
                    'vertical-align': 'middle', 'margin-top': margin_top}, className=className)


def baseline_img(app, is_long=True):
    """Returns a HTML element visually representing a baseline.
    The only parameter is a bool that defines if the representation
    is for a long baseline (True) or for a short baseline (False).
    """
    if is_long:
        baseline = html.Td(className='baseline-td-hr', children=html.Hr(className='hr-baseline'),
                           style={'width': '80%', 'padding': '0', 'margin': '0'})
    else:
        baseline = html.Td(className='baseline-td-hr', children=html.Hr(className='hr-baseline'),
                           style={'width': '30%', 'padding': '0', 'margin': '0'})

    return [html.Table(className='baseline-table', children=[
            html.Tr(className='baseline-tr', children=[
                html.Td(className='baseline-td-img', children=html.Img(width='17rem',
                        src=app.get_asset_url("icon-32.png"), alt='Antenna logo',
                        className='d-inline-block align-right img-baseline')),
                baseline,
                html.Td(className='baseline-td-img', children=html.Img(width='17rem',
                        src=app.get_asset_url("icon-32.png"), alt='Antenna logo',
                        className='d-inline-block align-left img-baseline')),
            ])])]


#################################################################################
# Some small graphical elements

def worldmap_plot(antennas):
    data = {"lat": [], "lon": [], "color": [], "symbol": [], "mode": "markers",
            "name": [], "text": [], "hovertemplate": []}
    for ant in antennas:
        data["lat"].append(ant.location.lat.value)
        data["lon"].append(ant.location.lon.value)
        data["color"].append('green')
        data["symbol"].append(0.1)
        data["name"].append(ant.name)
        data["text"].append(f"{ant.name}<br>({ant.country})<br> {ant.diameter}")
        data["hovertemplate"].append(f"{ant.name}<br>({ant.country})<br> {ant.diameter}")

    fig = px.scatter_geo(data,
                    lat="lat", lon="lon", text="name",
                    hover_name="text", hover_data=None)
    fig.update_layout(autosize=True, hovermode='closest', showlegend=False,
                      margin={'l': 0, 't': 0, 'b': 0, 'r': 0})

    return html.Div(className='row justify-content-center', children=dcc.Graph(figure=fig,
            responsive=True, className='fig-on-card',
                      config={'frameMargins': -10, 'showLink': False, 'displaylogo': False}))


def optimal_units(value, units):
    """Given a value (with some units), returns the unit choice from all
    `units` possibilities that better suits the value.
    It is meant for the following use:
    Given 0.02*u.Jy and units = [u.kJy, u.Jy, u.mJy, u.uJy], it will
    return 20*u.mJy.
    units should have a decreasing-scale of units, and all of them
    compatible with the units in `value`.
    """
    for a_unit in units:
        if 0.8 < value.to(a_unit).value <= 800:
            return value.to(a_unit)

    # Value too high or too low
    if value.to(units[0]).value > 1:
        return value.to(units[0])

    return value.to(units[-1])


#################################################################################
# Cards from the initial window where user keeps selecting options


def initial_window_start(app):
    """Very first window where the user can choose between going to the main window
    directly to manually selecting everything or going through the "wizard" guided mode.
    """
    return [
            html.Div(className='row card-deck justify-content-center text-center', children=[
                html.Button(id='button-initial-wizard', className='card-button btn btn-gray m-4',
                    title='Helps you to configure your observation in a step-by-step process.',
                    children=dbc.Card(dbc.CardBody([
                        html.H5("Guided mode", className='card-title'),
                        html.Img(height='80rem', src=app.get_asset_url('icon-newbie.png'),
                                 className='card-text m-3'),
                        html.Br(),
                        html.P("Easy step-by-step setup", className='card-text px-0')
                    ]), className='text-center shadow-0')
                ),
                html.Button(id='button-initial-expert', className='card-button btn btn-gray m-4',
                    title='Go to the main window containing all options to configure.',
                    children=[dbc.Card(dbc.CardBody([
                        html.H5("Manual mode", className='card-title'),
                        html.Img(height='80rem', src=app.get_asset_url('icon-expert.png'),
                                 className='card-text m-3'),
                        html.Br(),
                        html.P("Fully manual configuration", className='card-text px-0')
                    ]), className='text-center shadow-0')]
                )
            ])
        ]


def initial_window_pick_band():
    """Initial window with the introduction to the EVN Observation Planner and the band selection.
    """
    return [html.Div(className='row', children=[
                html.H3("Select the observing band"),
                html.P(["At which frequency or wavelength do you want to conduct your observation? "
                    "Note that, in any case, you will still be able to change your selection "
                    "afterwards in case you want to change to or compare different bands."])
            ], style={'text:align': 'justify !important'}),
            html.Br(),
            html.Div(className='justify-content-center', children=[html.Div(
                dcc.Slider(id='initial-band', min=0, max=len(fs.bands)-1,
                       value=tuple(fs.bands).index('18cm'), step=-1,
                       marks={i: fq for i,fq in enumerate(fs.bands)},
                       persistence=True,
                       updatemode='drag', included=False)), html.Br(),
                html.Div(id='initial-pickband-label', className='row justify-content-center', children=''),
                html.Br(),
                html.Div(className='row text-center',
                         children=html.Button('Continue', id='button-pickband',
                            className='btn btn-primary btn-lg'))
            ], style={'min-width': '33rem'}),
        ]


def initial_window_pick_network(app, vlbi_networks):
    """Initial window to introduce the default VLBI network(s) to be used.
    It will only allow the ones that can observe at the given wavelength.
    """
    return [html.Div(className='row', children=[
            html.H3('Select the VLBI Network(s)'),
            html.P(["Which network(s) will be used to observe your source? ", html.Br(),
                "This will select the default antennas from each network. Note that later "
                "you will be able to add or remove antennas as wished."])
            ]),
            html.Br(),
            html.Div(className='row justify-content-center', children=[
                network_card(app, a_network, vlbi_networks[a_network]['name'], "")
                for a_network in vlbi_networks if a_network != 'e-EVN'
            ]),
            html.Br(),
            html.Div(className='row justify-content-center',
                     children=html.Button('Continue', id='button-picknetwork',
                                disabled=True, className='btn btn-primary btn-lg')),
            # html.Div(hidden=True, children=[
            #     dcc.Dropdown(id='initial-array', options=[{'label': n, 'value': n} \
            #                  for n in vlbi_networks if n != 'e-EVN'], value=[],
            #                  persistence=True, multi=True)
            # ]),
            html.Div(style={'height': '20rem'})
        ]


def initial_window_pick_time():
    """Initial (second) window with the introduction to the EVN Observation Planner and
    the option to pick a specific observing time or let the tool to find them.
    """
    return [
        html.Div(className='row justify-content-center', children=[
            html.H3('Select the observing epoch and target'),
            html.P(["Introduce the epoch when the observation will be conducted and the target "
                "source to observe, if known,, or let the tool to find "
                "the most appropriate times for a given source/network."]),
            html.Br(),
            html.Div(className='row justify-content-center', children=[dbc.FormGroup([
                dbc.RadioItems(options=[{"label": "Do not specify source nor epoch",
                                         "value": SourceEpoch.UNSPECIFIED.value},
                                        {"label": "Find best epoch for the given source",
                                         "value": SourceEpoch.ONLY_SOURCE.value},
                                        {"label": "Define observing epoch and target source",
                                         "value": SourceEpoch.SOURCE_AND_EPOCH.value}],
                               value=SourceEpoch.SOURCE_AND_EPOCH.value, id="initial-timeselection",
                               inline=True, persistence=True),
                html.Br(),
                ], className='col-6', inline=True),
                html.Div(className='col-6', children=[
                    html.Small("", id='timeselection-div-smalltext', style={'color': '#999999'})
                ])]
            ),
            html.Br(),
            html.Div(className='col-7', children=[
                html.Div(id='initial-timeselection-div-source', className='row justify-content-center',
                    hidden=True, children=[
                            html.Label(['Source (name or coordinates)',
                                *tooltip(idname='popover-target',
                                     message="Source name or coordinates. " \
                                         "You may see an error if the given name is not properly resolved. "
                                         "J2000 coordinates are assumed in both forms: 00:00:00 00:00:00 or " \
                                         "00h00m00s 00d00m00s.")
                            ]),
                            html.Div(className='form-group', children=[
                                dcc.Input(id='initial-source', value=None, type='text',
                                          className='form-control', placeholder="hh:mm:ss dd:mm:ss",
                                          persistence=True),
                                html.Small(id='initial-error_source',
                                           className='form-text text-muted'),
                            ])
                    ]),
                    html.Div(id='initial-timeselection-div-epoch', hidden=False, children=[
                        html.Label([html.Br(), 'Start of observation (UTC)']),
                        *tooltip(idname='popover-startime', message="Select the date and "
                                    "time of the start of the observation (Universal, UTC, "
                                    "time). You will also see the day of the year (DOY) in "
                                    "brackets once the date is selected."),
                        html.Br(),
                        dcc.DatePickerSingle(id='initial-starttime', date=None, min_date_allowed=dt(1900, 1, 1),
                                             max_date_allowed=dt(2100, 1, 1),
                                             display_format='DD-MM-YYYY (DDD)',
                                             placeholder='Start date',
                                             first_day_of_week=1,
                                             initial_visible_month=dt.today(),
                                             persistence=True,
                                             className='form-picker'),
                        dcc.Dropdown(id='initial-starthour', placeholder="Start time (UTC)", value=None,
                                     options=[{'label': f"{hm//60:02n}:{hm % 60:02n}", \
                                               'value': f"{hm//60:02n}:{hm % 60:02n}"} \
                                              for hm in range(0, 24*60, 15)],
                                     persistence=True, className='form-hour'),
                        html.Small(id='initial-error_starttime', style={'color': 'red'},
                                   className='form-text text-muted')
                    ]),
                    html.Div(id='initial-timeselection-div-duration', className='row justify-content-center',
                        hidden=True, children=[
                            html.Label([html.Br(), 'Duration of the observation (in hours)']),
                            html.Div(className='form-group', children=[
                                dcc.Input(id='initial-duration', value=None, type='number', className='form-control',
                                       placeholder="Duration in hours", persistence=True, inputMode='numeric'),
                                html.Small(id='initial-error_duration', className='form-text text-danger')
                            ]),
                    ]),
    #             html.Div(id='initial-timeselection-div-guess', className='row justify-content-center',
    #                 children=[
    #                     html.Small("Choose this option if you just want to find out when your source will "
    #                         "be visible. It will pick the time range when more than 3 antennas can observe "
    #                         "simultaneously."),
    #                     html.Small("Note that this option may not provide the best (expected) results in the "
    #                         "case of combining different networks that are far apart (e.g. LBA together with "
    #                         "the EVN).")
    #             ]),
    #             html.Div(id='initial-timeselection-div-epoch', children=[
    #                 html.Label('Start of observation (UTC)'),
    #                 *tooltip(idname='popover-startime', message="Select the date and "
    #                             "time of the start of the observation (Universal, UTC, "
    #                             "time). You will also see the day of the year (DOY) in "
    #                             "brackets once the date is selected."),
    #                 html.Br(),
    #                 dcc.DatePickerSingle(id='initial-starttime', date=None, min_date_allowed=dt(1900, 1, 1),
    #                                      max_date_allowed=dt(2100, 1, 1),
    #                                      display_format='DD-MM-YYYY (DDD)',
    #                                      placeholder='Start date',
    #                                      first_day_of_week=1,
    #                                      initial_visible_month=dt.today(),
    #                                      persistence=True,
    #                                      className='form-picker'),
    #                 dcc.Dropdown(id='initial-starthour', placeholder="Start time (UTC)", value=None,
    #                              options=[{'label': f"{hm//60:02n}:{hm % 60:02n}", \
    #                                        'value': f"{hm//60:02n}:{hm % 60:02n}"} \
    #                                       for hm in range(0, 24*60, 15)],
    #                              persistence=True, className='form-hour'),
    #                 html.Small(id='initial-error_starttime', style={'color': 'red'},
    #                            className='form-text text-muted'),
    #                 html.Label('Duration of the observation (in hours)'),
    #                 html.Div(className='form-group', children=[
    #                     dcc.Input(id='initial-duration', value=None, type='number', className='form-control',
    #                                placeholder="Duration in hours", persistence=True, inputMode='numeric'),
    #                     html.Small(id='initial-error_duration',
    #                                className='form-text text-danger')
    #                 ])
    #             ])
        ]),
        html.Span(style={'height': '2rem'}),
        html.Div(className='row justify-content-center',
             children=html.Button('Continue', id='button-picktimes',
                        className='btn btn-primary btn-lg')),
    ])
    ]


def initial_window_pick_mode(app):
    """Initial window to select the observing mode: spectral line or continuum observation.
    """
    return [html.Div(className='row justify-content-center', children=[
                html.H3("What type of observation do you want?"),
                html.P(["This information will only be considered to pick up the default observing "
                    "setup that is more appropriate for you (in terms of bandwidth, number of spectral "
                    "channels, etc). You will still be able to tune these parameters afterwards."])
            ], style={'text:align': 'justify !important'}),
            html.Br(),
            html.Div(className='row card-deck justify-content-center text-center', children=[
                html.Button(id='button-mode-continuum', className='card-button btn btn-gray m-4',
                    title='Helps you to configure your observation in a step-by-step process.',
                    children=dbc.Card(dbc.CardBody([
                        html.H5("Continuum mode", className='card-title'),
                        html.Img(height='80rem', src=app.get_asset_url('waves.png'),
                                 className='card-text m-3'),
                        html.Br(),
                        html.P("Provides the maximum sensitivity", className='card-text px-0'),
                        html.P("Uses the maximum bandwidth available", className='card-text px-0')
                    ]), className='text-center shadow-0')
                ),
                html.Button(id='button-mode-line', className='card-button btn btn-gray m-4',
                            title='Go to the main window containing all options to configure.',
                            children=[dbc.Card(dbc.CardBody([
                                html.H5("Spectral line mode", className='card-title'),
                                html.Img(height='80rem', src=app.get_asset_url('waves-line.png'),
                                         className='card-text m-3'),
                                html.Br(),
                                html.P("Provides higher frequency resolution", className='card-text px-0'),
                                html.P("In general uses a reduced bandwidth", className='card-text px-0')
                            ]), className='text-center shadow-0')]
                )
            ]),
            html.Div(hidden=True, children=[dbc.Checklist(id='is_line',
                                                          options=[{'label': 'line obs', 'value': False}],
                                                          value=[])])
            ]


def initial_window_final():
    """The observing mode will be either 'cont' or 'line'.
    """
    return [html.Div(children=[
        # dcc.Tabs(id='tabs', value=''),
        html.Div(className='row justify-content-center', children=[
            html.H3('You are now ready'),
            html.P(["Press compute to produce the summary for your observation. "
                    "You would then see different tabs with the information. "
                    "You will also be able to change the setup and re-compute it."]),
            html.Br(),
        ]),
        html.Div(style={'height': '4rem'}),
        html.Div(className='row justify-content-center',
                 children=html.Button('Compute', id='antenna-selection-button',
                                      className='btn btn-primary btn-lg')),
        html.Br(),
        html.Div(className='col-9 text-center justify-content-center', children=[
            dcc.Loading(id="loading", children=[html.Div(id="loading-output",
                        className='text-center justify-content-center')],
                        type="dot"),
        ]),
        html.Div(className='row text-center justify-content-center',
                 children=dcc.ConfirmDialog(id='global-error', message=''))
        ])]





import asyncio
from typing import Optional, Union, Iterable, Sequence, Self
from importlib import resources
from functools import partial
from itertools import product
import subprocess
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import numpy as np
import yaml
from rich import print as rprint
# import scipy.ndimage
import datetime as dt
from enum import Enum, auto
from dataclasses import dataclass
from astropy import units as u
from astropy import coordinates as coord
from astropy.time import Time, TimeDelta
# TODO: uncomment this and move dependency outside observation.py
# import plotly.express as px
# from plotly.subplots import make_subplots

from .stations import Station, Stations
from .sources import Source, Scan, ScanBlock, SourceType, SourceNotVisible
from . import freqsetups


_NETWORKS = Stations.get_networks_from_configfile()
_STATIONS = Stations()


class Polarization(Enum):
    """Number of polarizations to be recorded. Possible values are:
    - SINGLE (only one polarization is recorded per antenna/baseline).
    - DUAL (the two RR,LL polarizations are recorded per antenna/baseline).
    - FULL (the full four RR,LL,RL,LR polarizations are recorded per antenna/baseline).
    """
    SINGLE = 1
    DUAL = 2
    FULL = 4

    def __str__(self):
        return self.name.lower()


class Observation(object):
    """Defines an observation of a single target source with a given network of stations.
    The observation can be set up in incremental steps (i.e. not all inputs are necessary
    at initialization time). Depending on the functions that you want to execute not all
    attributes may be required.
    An Observation allows the user to determine the elevation of the target source for
    each antenna in the network, the estimated rms noise level that would reach in such
    observation, or the resolution of the resulting images, assuming a standard neutral
    robust weighting.
    """

    def __init__(self, band: str, stations: Stations, scans: dict[str, ScanBlock],
                 times: Optional[Time] = None, duration: Optional[u.Quantity] = None,
                 datarate: Optional[Union[int, u.Quantity]] = None,
                 subbands: Optional[int] = None, channels: Optional[int] = None,
                 polarizations: Optional[Union[int, str]] = None,
                 inttime: Optional[Union[float, u.Quantity]] = None,
                 ontarget: float = 1.0, bits: int = 2):
        """Initializes an observation.
        Note that you can initialize an empty observation at this stage and add the
        information for the different attributes later. However, you may raise exception
        if running methods that require some of the unset attributes.

        Inputs
        - band : str
            Observing band to conduct the observation. Note that while this is a
            free-format string, it should match the type used when defining the bands
            at which each station can observe, and the default is the str format `XXcm`
            where XX represents the wavelength to observe in cm units.
            You can always check the available bands in `{Stations object}.observing_bands`.
        - stations : vlbiplanobs.stations.Stations
            Network of stations that will participate in the given observation.
        - scans : dict[str, ScanBlock]
            Scans to be observed during the observation.
            The key of the dict is just the name to refer to the specified ScanBlock as value.
        - times : astropy.time.Time
            An array of times defining the duration of the observation. That is,
            the first time will define the start of the observation and the last one
            will be the end of the observation. Note that the higher time resolution
            in `times` the more precise values you will obtain for antenna source
            visibility or determination of the rms noise levels. However, that will
            also imply a longer computing time. Steps of ~5-15 min seem appropriate
            for typical VLBI observations.
        - duration : astropy.units.Quantity
            Total duration of the observation. If 'times' provided, it will automatically be calculated
            from them. Otherwise it will be set from this value (e.g. for observations where the start
            time is not known).
        - datarate : int or astropy.units.Quantity
            Data rate for each antenna. It assumes that all antennas will run at the same
            data rate, which may not be true. If an int is introduce, it will be assumed to
            be given in Mbit/s. Otherwise, you can specify a Quantity with compatible units.
        - subbands : int
            Number of subbands in which the total bandwidth of the observation will be divided
            during correlation.
        - channels : int
            Number of channels for each subband to be created during correlation.
        - polarizations : int (1, 2, 4)  or str (single, dual, full)
            Number of polarizations to record in the observation. Three values are only possible
            0: single polarization.
            2: dual polarization (only direct terms: RR and LL, or XX and YY, are kept).
            4: full polarization (crossed terms are kept: RR, LL, RL, LR; or XX, YY, XY, YX).
        - inttime : float/int or astropy.units.Quantity
            Integration time (time resolution) that the final, correlated, data will show.
            If no units provided, seconds are assumed.
        - ontarget : float
            Fraction of the total observing time (end time minus start time) that will be
            spent on the target source. Note that in a typical VLBI observation only a fraction
            of the total observing time will end up in on-target source, commonly between 0.4-0.8
            (~40-80% of the time). This has an effect on the determination of the final rms noise level.
        - bits : int
            Number of bits at which the data have been recorded (sampled). A typical VLBI observation is
            almost always recorded with 2-bit data.
        """
        if isinstance(scans, dict) and all(isinstance(a_value, ScanBlock) for a_value in scans.values()):
            self.scans = scans
        else:
            raise ValueError("Scans needs to be either a ScanBlock or a list of ScanBlocks.")

        self._source_list: Optional[dict[str, Source]] = None
        # Because otherwise gstimes is not initialized
        if times is None:
            self._times = None
            self._gstimes = None
        else:
            self.times = times

        self.band = band
        self.datarate = datarate
        self.subbands = subbands
        self.channels = channels
        self.polarizations = polarizations if polarizations is not None else Polarization.FULL  # type: ignore
        self.inttime = inttime
        self._duration = duration
        self.stations = stations
        self.bitsampling = bits
        self.ontarget_fraction = ontarget
        self._uv_baseline = None
        self._uv_array = None
        self._rms = None
        self._synth_beam = None
        self._is_visible: Optional[dict[str, dict[str, coord.SkyCoord]]] = None
        self._is_always_visible: Optional[dict] = None
        self._when_visible: Optional[dict[str, list[Time | u.Quantity]]] = None
        self._altaz: Optional[dict] = None
        self._elevations: Optional[dict] = None

    @property
    def scans(self) -> Optional[dict[str, ScanBlock]]:
        """Returns the ScanBlock that will be observed during the observation.
        """
        return self._scans

    @scans.setter
    def scans(self, scans: Optional[dict[str, ScanBlock]]):
        self._scans = scans

    @property
    def sources(self) -> Optional[dict[str, Source]]:
        """Returns the sources included during the observation.
        It can return None if the target has not been set yet, showing a warning.
        """
        if self._scans is None:
            return None

        if self._source_list is not None:
            return self._source_list

        self._source_list = {}
        for a_scanblock in self._scans.values():
            for a_src in a_scanblock.sources():
                self._source_list[a_src.name] = a_src

        return self._source_list

    @property
    def sourcenames(self) -> Optional[list[str]]:
        """Returns the source names included during the observation
        It can return None if the target has not been set yet, showing a warning.
        """
        return [a_source.name for a_source in self.sources.values()] if self.sources is not None else None

    @property
    def times(self) -> Optional[Time]:
        """Returns the times when the observation runs as an astropy.time.Time object.
        It can return None if the times have not been set yet, showing a warning.
        """
        # if self._times is None:
        #     print("WARNING: 'times' not set yet but used in 'Observation'.")
        return self._times

    @times.setter
    def times(self, new_times: Optional[Time]):
        if (new_times is not None) and (not isinstance(new_times, Time)):
            raise ValueError("'times' must be an astropy.time.Time instance or be None")
        elif isinstance(new_times, Time) and new_times.size < 2:
            raise ValueError("'times' must have at least two time values (start and end of the observation).")

        self._times = new_times
        if self._times is None:
            self._gstimes = None
        else:
            self._gstimes = self._times.sidereal_time('mean', 'greenwich')

        self._uv_baseline = None
        self._uv_array = None
        self._rms = None
        self._synth_beam = None

    @property
    def gstimes(self) -> Optional[coord.angles.Longitude]:
        """Returns the GST times when the observation runs as an astropy.coordinates.angles.Longitude
        object (meaning in hourangle units).
        It can return None if the times have not been set yet, showing a warning.
        """
        return self._gstimes

    @property
    def duration(self) -> Optional[u.Quantity]:
        """Returns the total duration of the observation.
        """
        return self._duration if self.times is None else (self.times[-1] - self.times[0]).to(u.h)

    @property
    def band(self) -> str:
        """Returns the observing band at which the observation will be conducted.
        It can return None if the band has not been set yet, showing a warning.
        """
        return self._band

    @band.setter
    def band(self, new_band: str):
        if (new_band not in freqsetups.bands):
            raise ValueError("'new_band' needs to  match the following bands: "
                             f"{', '.join(freqsetups.bands.keys())} (wavelengths given in cm.)")

        self._band = new_band
        self._uv_baseline = None
        self._uv_array = None
        self._rms = None
        self._synth_beam = None

    @property
    def wavelength(self) -> u.Quantity:
        """Returns the central wavelength of the observation.
        """
        return float(self.band.replace('cm', ''))*u.cm

    @property
    def frequency(self) -> u.Quantity:
        """Returns the central frequency of the observations.
        """
        return 30*u.GHz/self.wavelength.to(u.cm).value

    @property
    def datarate(self) -> Optional[u.Quantity]:
        """Retuns the data rate (per station) used at which the observation is conducted.
        It can return None if the data rate has not been set yet, showing a warning.
        """
        return self._datarate

    @datarate.setter
    def datarate(self, new_datarate: Optional[u.Quantity]):
        """Sets the data rate used at each station during the observation.

        Inputs
        - new_datarate : astropy.units.Quantity  [e.g. Mb/s]
        """
        if new_datarate is None:
            self._datarate = None
        elif isinstance(new_datarate, int):
            if new_datarate <= 0:
                raise ValueError(f"datarate must be a positive number (currently {new_datarate})")

            self._datarate = new_datarate*u.Mbit/u.s
        elif isinstance(new_datarate, u.Quantity):
            if new_datarate <= 0:
                raise ValueError(f"datarate must be a positive number (currently {new_datarate})")

            self._datarate = new_datarate.to(u.Mbit/u.s)
        else:
            raise ValueError(f"Unknown type for datarate {new_datarate}"
                             "(int or astropy.units.Quantity (~bit/s) expected)")

        self._rms = None

    @property
    def subbands(self) -> Optional[int]:
        """Returns the number of subbands (also known as intermediate frequencies for AIPS users)
        in which the total bandwidth of the observation will be divided during correlation.
        It can return None if the number of subbands has not been set yet, showing a warning.
        """
        return self._subbands

    @subbands.setter
    def subbands(self, n_subbands: Optional[int]):
        if (n_subbands is not None) and (not isinstance(n_subbands, int)) and (n_subbands <= 0):
            raise ValueError("n_subbands needs to be a positive int, or None.")

        self._subbands = n_subbands

    @property
    def channels(self) -> Optional[int]:
        """Returns the number of channels in which each subband will be divided during correlation.
        It can return None if the number of channels has not been set yet, showing a warning.
        """
        return self._channels

    @channels.setter
    def channels(self, n_channels: Optional[int]):
        if (not isinstance(n_channels, int)) and (n_channels is not None) and (n_channels <= 0):
            raise ValueError("channels needs to be a positive int, or None.")

        self._channels = n_channels

    @property
    def polarizations(self) -> Polarization:
        """Returns the number of polarizations that will be stored in the final data.
        It can return None if the number of polarizations has not been set yet, showing a warning.
        """
        return self._polarizations

    @polarizations.setter
    def polarizations(self, pols: Union[int, str, Polarization]):
        if isinstance(pols, Polarization):
            self._polarizations = pols
            return

        match pols:
            case 1 | 'single':
                self._polarizations = Polarization.SINGLE
            case 2 | 'dual':
                self._polarizations = Polarization.DUAL
            case 4 | 'full':
                self._polarizations = Polarization.FULL
            case _:
                raise ValueError("Polarizations needs to be either a Polarization value, "
                                 "the number 1, 2, or 4, or the strings "
                                 "'single', 'dual' or 'full' (equivalent to the numbers, respectively).")

    @property
    def inttime(self) -> Optional[u.Quantity]:
        """Returns the integration time used when correlating the observation as an astropy.units.Quantity.
        It can return None if the integration time has not been set yet, showing a warning.
        """
        return self._inttime

    @inttime.setter
    def inttime(self, new_inttime: Optional[Union[int, float, u.Quantity]]):
        """Sets the integration time of the observation.
        Inputs
        - new_inttime float/int or astropy.units.Quantity.
            If no units provided, seconds are assumed.
        """
        if new_inttime is None:
            self._inttime = None
        elif isinstance(new_inttime, float) or isinstance(new_inttime, int):
            if new_inttime <= 0:
                raise ValueError(f"'inttime' must be a positive number (currently {new_inttime})")

            self._inttime = new_inttime*u.s
        elif isinstance(new_inttime, u.Quantity):
            if new_inttime <= 0:
                raise ValueError(f"'inttime' must be a positive number (currently {new_inttime})")

            self._inttime = new_inttime.to(u.s)
        else:
            raise ValueError(f"Unknown type for 'inttime' {new_inttime} (float/int/Quantity(~seconds) expected)")

    @property
    def ontarget_fraction(self) -> float:
        """Fraction of the total observing time spent on the target source.
        If scans include more sources than just the target source, then the fraction will be calculated
        from these scans, ignoring this fraction.
        """
        return self._ontarget

    @ontarget_fraction.setter
    def ontarget_fraction(self, ontarget: float):
        if not (0.0 < ontarget <= 1.0):
            raise ValueError("'ontarget_fraction' must be a float within (0.0, 1.0].")
        self._ontarget = ontarget
        self._rms = None

    @property
    def ontarget_time(self) -> Optional[u.Quantity]:
        """Total time spent on the target source during the observation.
        It can return None if the ontarget_fraction and duration have not been set yet, showing a warning.
        """
        return None if self.duration is None else self.duration*self.ontarget_fraction

    @property
    def bandwidth(self) -> Optional[u.Quantity]:
        """Returns the total bandwidth of the observation.
        It returns None if the attributes 'polarizations', 'datarate', or 'bitsampling'
        have not been set yet.
        """
        if None in (self.polarizations, self.datarate, self.bitsampling):
            return None

        pols = self.polarizations.value % 3 + self.polarizations.value // 3  # Either 1 or 2
        return (self.datarate/(pols*self.bitsampling*2)).to(u.MHz)  # type: ignore

    @property
    def bitsampling(self) -> int:
        """Returns the bit sampling at which the data are recorded during the observation.
        """
        return self._bitsampling

    @bitsampling.setter
    def bitsampling(self, new_bitsampling: int):
        """Sets the bit sampling of the observation.
        Inputs
        - new_bitsampling : int
            In bits.
        """
        if isinstance(new_bitsampling, int):
            self._bitsampling = new_bitsampling*u.bit
        else:
            raise ValueError(f"Unknown type for {new_bitsampling} (int, in bits, expected)")

    @property
    def stations(self) -> Stations:
        """Returns the network of stations 'Stations' that will participate in this observation
        observing the target source.
        """
        return self._stations

    @stations.setter
    def stations(self, new_stations: Stations):
        assert isinstance(new_stations, Stations)
        self._stations = new_stations
        self._uv_baseline = None
        self._uv_array = None
        self._rms = None
        self._synth_beam = None
        self._elevations = None
        self._is_visible = None
        self._is_always_visible = None
        self._altaz = None

    def sources_in_block(self, block_name: str, source_type: Optional[SourceType] = None) -> list[Source]:
        """Returns the sources that are included in a specific scan block.
        """
        if self._scans is None or block_name not in self._scans:
            raise ValueError(f"The ScanBlock named {block_name} could not be found.")

        return self._scans[block_name].sources(source_type)

    def sourcenames_in_block(self, block_name: str, source_type: Optional[SourceType] = None) -> list[str]:
        """Returns the sources that are included in a specific scan block.
        """
        if self._scans is None or block_name not in self._scans:
            raise ValueError(f"The ScanBlock named {block_name} could not be found.")

        return self._scans[block_name].sourcenames(source_type) if self._scans is not None else None

    def elevations(self) -> dict[str, dict[str, dict[str, coord.angles.Latitude]]]:
        """Returns the elevation of the sources for each stations participating in the observation
        for all the given observing times.

        Returns
            elevations : dict
                Dictionary where they keys are: the ScanBlock name, the source name,
                and then the station code names, and the values are the elevations at each time stamp.
        """
        if (self.sources is None) or (self.times is None):
            raise ValueError("The target and/or observing times have not been initialized")
        # the current timing checks provided (on a test observation):
        # Elevations with process: 16.56423762498889 s
        # Elevations with threads: 0.2133850830141455 s
        # Elevations with nothing: 0.36229008401278406 s
        if self._elevations is None:
            self._elevations = self._elevations_threads()

        return self._elevations

    # INFO: this code in the following was meant to test how the computation runs faster (process, threads, etc)
    def _elevation_func(self, a_scanblock: ScanBlock) -> dict[str, dict[str, Iterable]]:
        if (self.sources is None) or (self.times is None):
            raise ValueError("The target and/or observing times have not been initialized")

        # TODO: check if doing a multi-thread here will improve performance or not

        return {src: {a_station.codename: a_station.elevation(self.times, self.sources[src]) for a_station in self.stations}
                for src in self.sources}

    def _elevations_process(self) -> dict:
        """Returns the elevation of the target source for each stations participating in the observation
        for all the given observing times.

        It computes it using different processes in parallel.

        Returns
            elevations : dict
                Dictionary where they keys are the station code names, and the values will be
                an astropy.coordinates.angles.Latitude object with the elevation at each observing time.
        """
        with ProcessPoolExecutor() as pool:
            results = pool.map(self._elevation_func, self.stations)

        return {s.codename: r for s, r in zip(self.stations, results)}

    def _elevations_threads(self) -> dict:
        """Returns the elevation of the target source for each stations participating in the observation
        for all the given observing times.

        It computes it using different concurrent threads.

        Returns
            elevations : dict
                Dictionary where they keys are the station code names, and the values will be
                an astropy.coordinates.angles.Latitude object with the elevation at each observing time.
        """
        assert self.scans is not None
        with ThreadPoolExecutor() as pool:
            results = pool.map(self._elevation_func, self.scans.values())

        return {s: r for s, r in zip(self.scans.keys(), results)}

    def _elevations_single(self) -> dict:
        """Returns the elevation of the target source for each stations participating in the observation
        for all the given observing times.

        It computes it without parallelization.

        Returns
            elevations : dict
                Dictionary where they keys are the station code names, and the values will be
                an astropy.coordinates.angles.Latitude object with the elevation at each observing time.
        """
        elevations = {}
        for a_station in self.stations:
            elevations[a_station.codename] = a_station.elevation(self.times, self.sources)
        return elevations

    async def _elevations_async_func(self, a_station: Station) -> list:
        return a_station.elevation(self.times, self.sources)

    async def _elevations_async_launch(self) -> dict:
        tasks: dict = {}
        for station in self.stations:
            tasks[station.codename] = asyncio.create_task(self._elevations_async_func(station))
            await tasks[station.codename]

        return tasks

    def _elevations_asyncio(self) -> dict:
        return asyncio.run(self._elevations_async_launch())
        # return {s.codename: v for s,v in zip(self.stations, asyncio.run(self._elevations_async_launch()))}

    def altaz(self) -> dict[str, dict[str, coord.SkyCoord]]:
        """Returns the altitude/azimuth of the target source for each stations participating
        in the observation for all the given observing times.

        Returns
            altaz : dict
                Dictionary where they keys are the station code names, and the values will be
                another dictionary with the source name as keys, and
                an astropy.coordinates.SkyCoord object with the altitude and azimuth
                of the source at each observing time.
        """
        if None in (self.sources, self.times):
            raise ValueError("The target and/or observing times have not been initialized")

        if self._altaz is None:
            self._altaz = self._altaz_threads()

        return self._altaz

    def _altaz_threads(self) -> dict:
        assert self.sources is not None and self.scans is not None

        def compute_altaz_for_source(station_source: tuple) -> tuple:
            station, source = station_source[0], station_source[1]
            return station, source, station.altaz(self.times, source)

        temp: dict[str, dict[str, coord.SkyCoord]] = {station.codename: {} for station in self.stations}
        for ablockname, ablock in self.scans.items():
            temp[ablockname] = {source.name: {} for source in ablock.sources()}
            with ThreadPoolExecutor() as executor:
                results = list(executor.map(compute_altaz_for_source, [(station, source)
                                            for station in self.stations for source in ablock.sources()]))
            for (station, source, altaz) in results:
                temp[ablockname][source.name][station.codename] = altaz

        return temp

    def is_observable(self) -> dict[str, dict[str, list[bool]]]:
        """Returns whenever the given ScanBlock can be observed by each station for each time
        of the observation.

        Returns
            is_visible : dict
                Dictionary where they keys are the station code names, and the values will be
                a dictionary with the sources names as keys, and a tuple containing
                a numpy array with the indexes in the `Observation.times`
                array with the times where the target source can be observed by the station.

                In this sense, you can e.g. call obs.times[obs.is_visible[a_station_codename]]
                to get such times.
        """
        if None in (self.scans, self.times, self.sources):
            raise ValueError("The target, sources and/or observing times have not been initialized")

        if self._is_visible is None:
            def compute_is_visible_for_source(station_source: tuple) -> tuple:
                station, source = station_source[0], station_source[1]
                return station, source, station.is_observable(self.times, source)

            self._is_visible = {}
            for ablockname, ablock in self.scans.items():  # type: ignore
                with ThreadPoolExecutor() as executor:
                    results = list(executor.map(compute_is_visible_for_source, [(station, source)
                                                for station in self.stations
                                                for source in ablock.sources()]))  # type: ignore

                self._is_visible[ablockname] = {}
                for (station, source, visible) in results:
                    self._is_visible[ablockname][station.codename] = visible

        return self._is_visible

    def is_observable_at(self, time: Time) -> dict[str, dict[str, bool]]:
        """Returns whenever the given ScanBlock can be observed by each station at the given time.
        """
        def compute_is_visible_for_source(station_source: tuple) -> tuple:
            station, source = station_source[0], station_source[1]
            return station, source, station.is_observable(time, source)

        _is_visible_at = {}
        for ablockname, ablock in self.scans.items():  # type: ignore
            with ThreadPoolExecutor() as executor:
                results = list(executor.map(compute_is_visible_for_source, [(station, source)
                                            for station in self.stations
                                            for source in ablock.sources]))  # type: ignore

            _is_visible_at[ablockname] = {}
            for (station, source, visible) in results:
                _is_visible_at[ablockname][station.codename] = visible

        return _is_visible_at

    def can_be_observed(self) -> dict[str, dict[str, bool]]:
        """Returns whenever the sources can be observed by each station at least during part of the observation.

        Returns
            is_observable : dict
                Dictionary where they keys are the station code names, and the values will be
                a dictionary with the sources names as keys, and a boolean containing
                if the source can be observed by the station.
        """
        return {s: {k: any(v) for k, v in i.items()} for s, i in self.is_observable().items()}

    def is_always_observable(self) -> dict[str, dict[str, bool]]:
        """Returns whenever the target source can be observed by each station along the whole observation.

        Returns
            is_visible : dict
                Dictionary where they keys are the station code names, and the values will be
                a tuple containing a numpy array with the boolean indicating if the given BlockScan
                (same order as in Observation.sources) is always visible for the given station.
        """
        if None in (self.scans, self.times, self.sources):
            raise ValueError("The target, sources and/or observing times have not been initialized")

        if self._is_always_visible is None:
            def compute_is_always_visible_for_source(station_source: tuple) -> tuple:
                station, source = station_source[0], station_source[1]
                return station, source, station.is_always_observable(self.times, source)

            self._is_always_visible = {ablockname: {} for ablockname in self.scans}  # type: ignore
            for ablockname, ablock in self.scans.items():  # type: ignore
                with ThreadPoolExecutor() as executor:
                    results = list(executor.map(compute_is_always_visible_for_source, [(station, source)
                                                for station in self.stations
                                                for source in ablock.sources()]))  # type: ignore

                for (station, source, visible) in results:
                    self._is_always_visible[ablockname][station.codename] = visible

        return self._is_always_visible

    def when_is_observable(self, min_stations: int = 3,
                           mandatory_stations: Optional[str | list[str]] = None,
                           stations_all_time: bool = False, within_time_range: Optional[Time] = None,
                           return_gst: bool = False) -> dict[str, list[Time | u.Quantity]]:
        """Returns the time range when the different BlockScans/sources are visible verifying the set
        requirements.
        If the time of the observation is set, it will check if the source is observable only within that
        time range, otherwise it will check at any possible time.

        Inputs
        - min_stations : int  [default = 3]
            Minimum number of stations that are allowed to consider the source as observable.
        - mandatory_stations : list[str]  [default = None]
            Defines the stations that must be able to observe the source to be consider for observations.
            It must be a list of strings with the codes or names of the stations.
            The wildcard 'all' is possible, indicating that only times where all antennas can observe
            the source are allowed.
        - stations_all_time : bool  [default = False]
            If all stations must be observing the source at all times or this is not required.
        - within_time_range : Time  [default = None]
            Time range within which the source is observable.
            Must contain only two times (start and end time).
        - return_gst : bool  [default = False]
            Defines if the returned times for the start and end of the observation should be in GST time
            instead of UTC.

        Returns
            - (t0, t1, ...)  : tuple with two astropy.units.Quantity[Longitude] objects.
                The start and end (GST) time of the same period of time when the source is visible by
                enough stations.
                If 'return_gst' is False, the tuple will contain two astropy.time.Time objects instead,
                with the UTC times for the start and end times.
                It will be always an even number of entries. If more than two, means that there are
                multiple time ranges where it can be observed following t0, t1, t2, t3, ...
                so the source can be observed between t0 and t1, t2 and t3, etc.

        Exceptions
        - It may raise the exception SourceNotVisible if the target source is not visible by
          enough stations.
        """
        if self.sources is None:
            raise ValueError("The sources have not been initialized")

        if self._is_visible is None:
            _ = self.is_observable()
        # if self._when_visible is not None:
        #     if return_gst:
        #         return {s: list([t.sidereal_time('mean', 'greenwich') for t in self._when_visible[s]])
        #                 for s in self._when_visible}
        #
        #     return self._when_visible

        if within_time_range is None:
            if self.times is None:
                # Using the autonm equinox as that one GST ~= UTC
                within_time_range = Time('2025-09-21', scale='utc') + np.arange(0, 2, 0.01)*u.day
            else:
                within_time_range = self.times
        else:
            assert len(within_time_range) == 2, \
                "The time range must contain only two times (start and end time)"
            assert within_time_range[1] <= within_time_range[0], \
                "The end time must be larger than the start time"
            within_time_range = within_time_range[0] + \
                np.arange((within_time_range[1] - within_time_range[0]).to(u.day).value, 0.01)*u.day

        if mandatory_stations == 'all':
            mandatory_stations = self.stations.station_codenames
            min_stations = len(mandatory_stations)

        # print(f"times: {within_time_range}\nmandatory: {mandatory_stations}")

        result: dict[str, list[Time | u.Quantity]] = {}
        for blockname, station_visibility in self._is_visible.items():
            # visible_times = [time for i, time in enumerate(within_time_range)
            #                  if ((sum(visibility[station][i] for station in visibility) >= min_stations)
            #                      and ((mandatory_stations is None)
            #                           or all(visibility[station][i] for station in mandatory_stations)))]
            visible_times = [False for i in range(len(within_time_range))]
            for i, time in enumerate(within_time_range):
                if ((sum(vis[i] for vis in station_visibility.values()) >= min_stations)
                    and ((mandatory_stations is None)
                         or all(station_visibility[station][i] for station in mandatory_stations))):
                    visible_times[i] = True

            # print(f"{visible_times=}  (total of {len(visible_times)} timestamps)")
            # print(f"True values at {', '.join([str(i) for i, v in enumerate(visible_times) if v])}")
            diff = [i + 1*(not value) for i, value in enumerate(visible_times)
                    if value != (visible_times + [False])[i+1] or (i == 0 and value)]
            # print(f"{diff=}")
            result[blockname] = list(zip(within_time_range[diff[::2]], within_time_range[diff[1::2]]))
            # print(f"{result=}")

        self._when_visible = result
        if return_gst:
            return {s: list([t.sidereal_time('mean', 'greenwich') for t in result[s]]) for s in result}

        return self._when_visible

    def scheduler(self) -> list[Scan]:
        """For the given observing time and scan blocks, it will schedule in an optimal way the scans
        for the different sources along the observation.
        """
        raise NotImplementedError



    def longest_baseline(self) -> tuple[str, u.Quantity]:
        """Returns the longest baseline in the observation for each source.

        Returns
        - 'source name': ('{ant1}-{ant2}', length) : dict[str, tuple[str, astropy.units.Quantity]]
            - '{ant1}-{ant2}' : str
                Composed by the codenames of the two antennas (ant1, ant2) conforming the longest baseline.
            - length : astropy.units.Quantity
                The projected length of the baseline as seen from the target source position.
        """
        uv = self.get_uv_baseline()
        longest_bl = {'bl': '', 'value': 0.0*u.m}
        for a_bl in uv:
            bl_length = np.sqrt(np.max((uv[a_bl]**2).sum(axis=1)))
            if bl_length > longest_bl['value']:
                longest_bl['bl'] = a_bl
                longest_bl['value'] = bl_length

        return longest_bl['bl'], longest_bl['value']*self.wavelength

    def shortest_baseline(self) -> tuple:
        """Returns the shortest baseline in the observation.

        Returns
        - ('{ant1}-{ant2}', length) : tuple
            - '{ant1}-{ant2}' : str
                Composed by the codenames of the two antennas (ant1, ant2) conforming the shortest baseline.
            - length : astropy.units.Quantity
                The projected length of the baseline as seen from the target source position.
        """
        uv = self.get_uv_baseline()
        shortest_bl = {'bl': '', 'value': 0.0*u.m}
        for a_bl in uv:
            bl_length = np.sqrt(np.max((uv[a_bl]**2).sum(axis=1)))
            if (shortest_bl['value'] == 0.0*u.m) or (bl_length < shortest_bl['value']):
                shortest_bl['bl'] = a_bl
                shortest_bl['value'] = bl_length

        return shortest_bl['bl'], shortest_bl['value']*self.wavelength

    def bandwidth_smearing(self) -> u.Quantity:
        """Returns the bandwidth smearing expected for the given observation.

        The peak response to a point target source decreases at positions farther away from the
        pointing (correlated) sky position due to the frequency averaging performed in the data.

        This function returns the angular separation at which the bandwidth smearing produces
        a reduction of a 10% in the response of the telescope. The field of view should then
        be limited to this range to avoid significant loses.
        """
        return ((49500*u.arcsec*u.MHz*u.km)*self.channels /
                (self.longest_baseline()[1]*self.bandwidth/self.subbands)).to(u.arcsec)

    def time_smearing(self) -> u.Quantity:
        """Returns the time smearing expected for the given observation.

        The peak response to a point target source decreases at positions farther away from the
        pointing (correlated) sky position due to the time averaging performed in the data.

        This function returns the angular separation at which the time smearing produces
        a reduction of a 10% in the response of the telescope. The field of view should then
        be limited to this range to avoid significant loses.
        """
        return ((18560*u.arcsec*u.km*u.s/u.cm) *
                (self.wavelength/(self.longest_baseline()[1]*self.inttime))).to(u.arcsec)

    def datasize(self) -> Optional[u.Quantity]:
        """Returns the expected size for the output FITS IDI files.

        A regular observation with the European VLBI Network is stored in FITS IDI files,
        typically several 2-GB files. This function provides an estimation of the total
        size for these stored files.
        Note that this function does not take into account down times for the different
        stations. The provided value will thus always be un upper-limit for the real, final,
        value.
        """
        if None in (self.duration, self.inttime, self.subbands, self.channels):
            return None

        temp = len(self.stations)**2 * (self.duration/self.inttime).decompose()  # type: ignore
        temp *= self.polarizations.value*self.subbands*self.channels  # type: ignore
        return temp*1.75*u.GB/(131072*3600)

    # TODO: re-write this function assuming all sources, and after the observation has been set
    def thermal_noise(self) -> u.Quantity:
        """Returns the expected rms thermal noise for the given observation.

        Each antenna has a different sensitivity for the observing band (established from
        the SEFD value). This function computes the available baselines at each timestamp
        and estimates the final thermal noise reached when integrating the whole observation.

        Note that this function takes into account when each station can observe the source,
        but does not take into account sensitivity drops doe to external factors like e.g.
        low elevations of the source. The provided thermal noise is also assumed when a natural
        weighting is applied to the data when imaging. The thermal noise can thus be a bit
        higher if an uniform roubst is used.

        If the source has not been set, it will assume that all antennas observe all time.
        """
        raise NotImplementedError
        if self._rms is not None:
            return self._rms

        if self.target is None:
            sefds = [stat.sefd(self.band) for stat in self.stations]
            dt = (self.times[-1]-self.times[0]).to(u.s).value
            temp = 0.0
            for j in range(len(sefds)):
                for k in range(j+1, len(sefds)):
                    temp += dt/(sefds[j]*sefds[k])
        else:
            main_matrix = np.zeros((len(self.times), len(self.stations)))
            visible = self.is_visible()
            for i, stat in enumerate(self.stations):
                main_matrix[:, i][visible[stat.codename]] = stat.sefd(self.band)
            # Determines the noise level for each time stamp.
            temp = 0.0
            for i, ti in enumerate(main_matrix[:-1]):
                sefds = ti[np.where(ti > 0.0)]
                for j in range(len(sefds)):
                    for k in range(j+1, len(sefds)):
                        temp += (self.times[i+1]-self.times[i]).to(u.s).value/(sefds[j]*sefds[k])

            if np.abs(temp - 0.0) < 1e-5:
                # No sources visible
                raise SourceNotVisible('No single baseline can observe the source.')

        temp = 1.0/np.sqrt(temp*self.ontarget_fraction)
        self._rms = ((1.0/0.7)*temp/np.sqrt(self.datarate.to(u.bit/u.s).value/2))*u.Jy

        return self._rms

    def get_uv_baseline(self) -> dict[str, dict[str, u.Quantity]]:
        """Returns the (u, v) values for each baseline and each timestamp for which the source
        is visible.

        Returns
        - {source name: '{ant1}-{ant2}': uv_data} : dict
            - '{ant1}-{ant2}' : str
                The keys of the dictionary are the baselines forming the array, where ant1 and ant2
                are the code names of each station.
            - uv_data : astropy.units.Quantity
                A 2-d array with all (u, v) values for each timestamp in `times` when the source is
                visible for the given baseline. (u,v) are given in lambda units.
                Note that complex conjugate values are not provided.

        Exceptions
        - It may raise the exception SourceNotVisible if no baselines can observe the source
          at all during the observation.
        """
        if self._uv_baseline is not None:
            return self._uv_baseline

        bl_uv_up = {}

        def compute_per_source(hourangles, target):
            nstat = len(self.stations)
            # Determines the xyz of all baselines. Time independent
            bl_xyz = np.empty(((nstat*(nstat-1))//2, 3))
            bl_names = []
            s = [ant.location for ant in self.stations]
            for i in range(nstat):
                for j in range(i+1, nstat):
                    # An unique number defining a baseline
                    k = int(i*(nstat-1) - sum(range(i)) + j-i)
                    bl_xyz[k-1, :] = np.array([ii.value for ii in s[i].to_geocentric()]) - \
                        np.array([ii.value for ii in s[j].to_geocentric()])
                    bl_names.append("{}-{}".format(self.stations[i].codename,
                                                   self.stations[j].codename))

            # Matrix to convert xyz to uvw for each timestamp (but w is not considered)
            if self.target is None:
                m = np.array([[np.sin(hourangle), np.cos(hourangle), np.zeros(len(hourangle))],
                              [-np.sin(45*u.deg)*np.cos(hourangle),
                               np.sin(45*u.deg)*np.sin(hourangle),
                               np.cos(45*u.deg)*np.ones(len(hourangle))]])
            else:
                m = np.array([[np.sin(hourangle), np.cos(hourangle), np.zeros(len(hourangle))],
                              [-np.sin(self.target.dec)*np.cos(hourangle),
                               np.sin(self.target.dec)*np.sin(hourangle),
                               np.cos(self.target.dec)*np.ones(len(hourangle))]])

            bl_uv = np.array([m[:, :, i] @ bl_xyz.T for i in range(m.shape[-1])])*u.m

            if self.target is None:
                for i, bl_name in enumerate(bl_names):
                    ant1, ant2 = bl_name.split('-')
                    bl_uv_up[bl_name] = (bl_uv[:, :, i] / self.wavelength).decompose()

                self._uv_baseline = bl_uv_up
                return bl_uv_up
            else:
                ants_up = self.is_visible()
                for i, bl_name in enumerate(bl_names):
                    ant1, ant2 = bl_name.split('-')
                    bl_up = (np.array([a for a in ants_up[ant1][0] if a in ants_up[ant2][0]]), )
                    if len(bl_up[0]) > 0:
                        bl_uv_up[bl_name] = (bl_uv[:, :, i][bl_up] / self.wavelength).decompose()

                if len(bl_uv_up.keys()) == 0:
                    raise SourceNotVisible

                self._uv_baseline = bl_uv_up


        if self.sources is None:
            # Just assumes a source at +/-45 deg elevation without taking into account the Earth
            hourangle = self.gstimes
            # print("WARNING: 'target' is not set yet but used in 'Observation'")
        else:
            hourangle = (self.gstimes - self.target.ra.to(u.hourangle)).value % 24*u.hourangle

        nstat = len(self.stations)
        # Determines the xyz of all baselines. Time independent
        bl_xyz = np.empty(((nstat*(nstat-1))//2, 3))
        bl_names = []
        s = [ant.location for ant in self.stations]
        for i in range(nstat):
            for j in range(i+1, nstat):
                # An unique number defining a baseline
                k = int( i*(nstat-1) - sum(range(i)) + j-i )
                bl_xyz[k-1,:] = np.array([ii.value for ii in s[i].to_geocentric()]) - \
                             np.array([ii.value for ii in s[j].to_geocentric()])
                bl_names.append("{}-{}".format(self.stations[i].codename,
                                               self.stations[j].codename))

        # Matrix to convert xyz to uvw for each timestamp (but w is not considered)
        if self.target is None:
            m = np.array([[np.sin(hourangle), np.cos(hourangle), np.zeros(len(hourangle))],
                      [-np.sin(45*u.deg)*np.cos(hourangle),
                      np.sin(45*u.deg)*np.sin(hourangle),
                      np.cos(45*u.deg)*np.ones(len(hourangle))]])
        else:
            m = np.array([[np.sin(hourangle), np.cos(hourangle), np.zeros(len(hourangle))],
                      [-np.sin(self.target.dec)*np.cos(hourangle),
                      np.sin(self.target.dec)*np.sin(hourangle),
                      np.cos(self.target.dec)*np.ones(len(hourangle))]])

        bl_uv = np.array([m[:,:,i] @ bl_xyz.T  for i in range(m.shape[-1])])*u.m

        if self.target is None:
            for i,bl_name in enumerate(bl_names):
                ant1, ant2 = bl_name.split('-')
                bl_uv_up[bl_name] = (bl_uv[:,:,i]/self.wavelength).decompose()

            self._uv_baseline = bl_uv_up
            return bl_uv_up
        else:
            ants_up = self.is_visible()
            for i,bl_name in enumerate(bl_names):
                ant1, ant2 = bl_name.split('-')
                bl_up = (np.array([a for a in ants_up[ant1][0] if a in ants_up[ant2][0]]), )
                if len(bl_up[0]) > 0:
                    bl_uv_up[bl_name] = (bl_uv[:,:,i][bl_up]/self.wavelength).decompose()

            if len(bl_uv_up.keys()) == 0:
                raise SourceNotVisible

            self._uv_baseline = bl_uv_up
            return bl_uv_up


    def get_uv_array(self) -> np.ndarray:
        """Returns the (u, v) values for each baseline and each timestamp for which the source
        is visible.

        The difference with `get_uv_baseline` is that `get_uv_array` only returns the (u,v)
        values, dropping the information of baselines and times to which these values belong to.

        Returns a (N, 2)-dimensional numpy.ndarray containing all N (u,v) points resulting for
        each timestamp and baseline. The (u,v) values are given in lambda units.
        Note that complex conjugate values are not provided.

        Exceptions
        - It may raise the exception SourceNotVisible if no baselines can observe the source
          at all during the observation.
        """
        if self._uv_array is not None:
            return self._uv_array

        bl_uv_up = self.get_uv_baseline()
        tot_length = 0
        for bl_name in bl_uv_up:
            tot_length += bl_uv_up[bl_name].shape[0]

        uvvis = np.empty((tot_length, 2))
        i = 0
        for bl_name in bl_uv_up:
            uvvis[i:i+bl_uv_up[bl_name].shape[0],:] = bl_uv_up[bl_name]
            i += bl_uv_up[bl_name].shape[0]
        self._uv_array = uvvis
        return self._uv_array


    def synthesized_beam(self) -> dict:
        """Estimates the resulting synthesized beam of the observations based on
        the expected (u,v) coverage.

        This is just an estimation made by a ellipse fitting to the (u, v) coverage,
        from which we obtain the resolution on the two axes following
        https://science.nrao.edu/facilities/vlba/docs/manuals/oss/ang-res
            theta_HPBW (mas)  \\sim 2063 x lambda(cm)/b_max^km

        Note that different robust weighting maps during imaging would provide slightly
        different synthesized beams. The provided value here does not assumed any weighting
        in the data. A natural weighting would thus likely provide a slightly larger beam,
        while an uniform weighting would provide a slightly smaller beam.

        Returns a dict with the following keys: 'bmaj', 'bmin', 'pa' (major axis, minor axis,
        and position angle).
        The three values are astropy.units.Quantity objects with units of angles.
        """
        if self._synth_beam is not None:
            return self._synth_beam

        resolution = lambda bl : ((2.063e8*u.mas)/bl).to(u.mas)
        uvvis = self.get_uv_array()
        # Transform the uv points into r,theta (polar) points
        uvvis_polar = np.empty_like(uvvis)
        uvvis_polar[:,0] = np.sqrt((uvvis**2).sum(axis=1)) # radius
        uvvis_polar[:,1] = np.arctan2(uvvis[:,1], uvvis[:,0]) # theta
        # Defines the BMAJ and PA
        bl_bmaj = np.max(uvvis_polar[:,0])
        bl_bmaj_theta = uvvis_polar[:,1][np.where(uvvis_polar[:,0] == bl_bmaj)][0]
        # Gets the BMIN and an orthogonal projection
        bl_bmin_theta = ( bl_bmaj_theta + np.pi/2 ) % (2*np.pi)
        bl_bmin = np.max(np.abs(uvvis.dot(np.array([np.cos(bl_bmin_theta),
                                                    np.sin(bl_bmin_theta)]))))

        self._synth_beam = {'bmaj': resolution(bl_bmin), 'bmin': resolution(bl_bmaj),
                            'pa': (bl_bmaj_theta*u.rad).to(u.deg)}
        return self._synth_beam


    def get_dirtymap(self, pixsize: int = 1024, robust: str = "natural", oversampling: int = 4):
        """Returns the dirty beam produced for the given observation.

        Input:
            - pixsize : int
                Size in pixels of the returned (squared) image. By default 1024.
            - robust : str
                The weighting Briggs robust used to compute the dirty map.
                It must be either 'natural' or 'uniform'.
            - oversampling : int
                Oversampling factor when plotting the dirty map.
                Recommended values are between 1 and 10.
                CURRENTLY THIS PARAMETER IS IGNORED. WILL BE IMPLEMENTED IN A LATER VERSION

        Returns:
            - dirty_image : (pixsize x pixsize) np.array
                The dirty image in intensity.
            - laxis : np.array
                An array representing the values of each pixel in the image in mas for each axis.

        """
        assert robust in ('natural', 'uniform')
        assert isinstance(pixsize, int)
        # assert isinstance(oversampling, int) and 20 > oversampling >= 1
        # Oversampling the dirty beam
        uvimg = np.zeros((pixsize, pixsize))

        # Gridding uv.  Inspired from VNSIM code (https://github.com/ZhenZHAO/VNSIM)
        # uvdata = self.get_uv_array()
        # uvscaling = pixsize/(2.0*1.025*np.max(np.max(np.abs(uvdata), axis=0)))
        # uvimg[pixsize//2 + np.trunc(uvdata[:,0]*uvscaling).astype(int), \
        #       pixsize//2 + np.trunc(uvdata[:,1]*uvscaling).astype(int)] += 1
        #
        # # Recovering the requested size for the image
        # # if oversampling > 1:
        # #     uvimg = scipy.ndimage.zoom(uvimg, oversampling, order=1)
        # # uvimg = uvimg[int(pixsize*0.5):int(pixsize*1.5), int(pixsize*0.5):int(pixsize*1.5)]
        # if robust == 'uniform':
        #     if np.max(uvimg) > 1:
        #         uvimg[np.where(uvimg > 0)] = 1
        #     else:
        #         print('\n\nWARNING: Uniform dirty map same as natural one.\n\n')
        #
        # dirty_beam = np.real(np.fft.ifftshift(np.fft.ifft2(np.fft.fftshift(uvimg/np.max(uvimg)))))
        # if oversampling > 1:
        #     dirty_beam = scipy.ndimage.zoom(dirty_beam, oversampling, order=1)
        # imgsize = (uvscaling*u.rad/(2*oversampling)).to(u.mas) # angular equivalent size of the resulting image
        # return dirty_beam[int(pixsize*(oversampling-1)/2):int(pixsize*(oversampling+1)/2), \
        #                   int(pixsize*(oversampling-1)/2):int(pixsize*(oversampling+1)/2)].T/np.max(dirty_beam), \
        #        np.linspace(-imgsize, imgsize, pixsize)
        #         # return dirty_beam.T/np.max(dirty_beam), \
        #

        oversampling = 1
        # Gridding uv.  Inspired from VNSIM code (https://github.com/ZhenZHAO/VNSIM)
        uvdata = self.get_uv_array()
        uvscaling = (oversampling*pixsize)/(2*1.05*np.max(np.max(np.abs(uvdata), axis=0)))
        if robust == 'natural':
            uvimg[oversampling*pixsize//2 + np.trunc(uvdata[:,0]*uvscaling).astype(int), \
                  oversampling*pixsize//2 + np.trunc(uvdata[:,1]*uvscaling).astype(int)] += 1
        else:
            uvimg[oversampling*pixsize//2 + np.trunc(uvdata[:,0]*uvscaling).astype(int), \
                  oversampling*pixsize//2 + np.trunc(uvdata[:,1]*uvscaling).astype(int)] = 1

        # Recovering the requested size for the image
        # uvimg = uvimg[int(pixsize*0.5):int(pixsize*1.5), int(pixsize*0.5):int(pixsize*1.5)]
        dirty_beam = np.real(np.fft.ifftshift(np.fft.ifft2(np.fft.fftshift(uvimg/np.max(uvimg)))))
        imgsize = (uvscaling*u.rad/2).to(u.mas) # angular equivalent size of the resulting image
        return dirty_beam[int(pixsize*(oversampling-1)/2):int(pixsize*(oversampling+1)/2), \
                          int(pixsize*(oversampling-1)/2):int(pixsize*(oversampling+1)/2)].T/np.max(dirty_beam), \
               np.linspace(-imgsize, imgsize, pixsize)



    def print_obs_times(self, date_format='%d %B %Y'):
        """Returns the time range (starttime-to-endtime) of the observation in a smart way.
        If the observation lasts for less than one day it omits the end date:
                20 January 1971
                10:00-20:00 UTC
                GST range: 05:00-15:00

        If the observation ends the day after, then it returns:
                20 January 1971
                10:00-20:00 UTC (+1d)
                GST range: 05:00-15:00

        If the observation is longer, then it returns
                20 January 1971 10:00 to  24 January 1971 20:00 UTC
                GST range: 05:00-15:00

        Input:
            - date_format : str [OPTIONAL]
                Format for the date part (only the date part, not the times) of
                the string to represent the time range.
        Output:
            - strtime : str
                A string showing the time-range of the observation.
        """
        if self.gstimes is None:
            return "Observing time unspecifed."
        gsttext = "{:02n}:{:02.2n}-{:02n}:{:02.2n}".format((self.gstimes[0].hour*60) // 60,
                                                  (self.gstimes[0].hour*60) % 60,
                                                  (self.gstimes[-1].hour*60) // 60,
                                                  (self.gstimes[0].hour*60) % 60)
        if self.times[0].datetime.date() == self.times[-1].datetime.date():
            return "{} {}-{} UTC\nGST range: {}".format(self.times[0].datetime.strftime(date_format),
                                        self.times[0].datetime.strftime('%H:%M'),
                                        self.times[-1].datetime.strftime('%H:%M'), gsttext)
        elif (self.times[-1] - self.times[0]) < 24*u.h:
            return "{} {}-{} UTC (+1d)\nGST range: {}".format(
                                        self.times[0].datetime.strftime(date_format),
                                        self.times[0].datetime.strftime('%H:%M'),
                                        self.times[-1].datetime.strftime('%H:%M'), gsttext)
        else:
            return "{} {} to {} {} UTC\nGST range: {}".format(
                                        self.times[0].datetime.strftime(date_format),
                                        self.times[0].datetime.strftime('%H:%M'),
                                        self.times[-1].datetime.strftime(date_format),
                                        self.times[-1].datetime.strftime('%H:%M'), gsttext)


    def get_fig_ant_elev(self):
        data_fig = []
        data_dict = self.elevations()
        # Some reference lines at low elevations
        for ant in data_dict:
            data_fig.append({'x': self.times.datetime, 'y': data_dict[ant].value,
                            'mode': 'lines', 'hovertemplate': "Elev: %{y:.2n}º<br>%{x}",
                            'name': self.stations[ant].name})

        data_fig.append({'x': self.times.datetime, 'y': np.zeros_like(self.times)+10,
                         'mode': 'lines', 'hoverinfo': 'skip', 'name': 'Elev. limit 10º',
                         'line': {'dash': 'dash', 'opacity': 0.5, 'color': 'gray'}})
        data_fig.append({'x': np.unwrap(self.gstimes.value*2*np.pi/24)*24/(2*np.pi), 'y': np.zeros_like(self.times)+20,
                         'xaxis': 'x2', 'mode': 'lines', 'hoverinfo': 'skip',
                         'name': 'Elev. limit 20º', 'line': {'dash': 'dot', 'opacity': 0.5,
                         'color': 'gray'}})
        return {'data': data_fig,
                'layout': {'title': 'Source elevation during the observation',
                           'hovermode': 'closest',
                           'xaxis': {'title': 'Time (UTC)', 'showgrid': False,
                                     'ticks': 'inside', 'showline': True, 'mirror': False,
                                     'hovermode': 'closest', 'color': 'black'},
                           'xaxis2': {'title': {'text': 'Time (GST)', 'standoff': 0},
                                      'showgrid': False, 'overlaying': 'x', #'dtick': 1.0,
                                      'tickvals': np.arange(np.ceil(self.gstimes.value[0]),
                                            np.floor(np.unwrap(self.gstimes.value*2*np.pi/24)[-1]*24/(2*np.pi))+1),
                                      'ticktext': np.arange(np.ceil(self.gstimes.value[0]),
                                            np.floor(np.unwrap(self.gstimes.value*2*np.pi/24)[-1]*24/(2*np.pi))+1)%24,
                                      'ticks': 'inside', 'showline': True, 'mirror': False,
                                      'hovermode': 'closest', 'color': 'black', 'side': 'top'},
                           'yaxis': {'title': 'Elevation (degrees)', 'range': [0., 92.],
                                     'ticks': 'inside', 'showline': True, 'mirror': "all",
                                     'showgrid': False, 'hovermode': 'closest'},
                           'zeroline': True, 'zerolinecolor': 'k'}}




    def get_fig_ant_up(self):
        data_fig = []
        data_dict = self.is_visible()
        gstimes = np.unwrap(self.gstimes.value*2*np.pi/24)*24/(2*np.pi)
        gstimes = np.array([dt.datetime(self.times.datetime[0].year, self.times.datetime[0].month,
                            self.times.datetime[0].day) + dt.timedelta(seconds=gst*3600) for gst in gstimes])
        for i,ant in enumerate(data_dict):
            # xs = [obs.times.datetime[0].date() + datetime.timedelta(seconds=i*3600) for i in np.unwrap(obs.gstimes.value*2*np.pi/24)[data_dict[ant]]*24/(2*np.pi)]
            xs = gstimes[data_dict[ant]]
            data_fig.append({'x': xs,
                             'y': np.zeros_like(data_dict[ant][0])-i, 'type': 'scatter',
                             'hovertemplate': "GST %{x}",
                             'mode': 'markers', 'marker_symbol': "41",
                             'hoverinfo': "skip",
                             'name': self.stations[ant].name})

        data_fig.append({'x': self.times.datetime, 'y': np.zeros_like(self.times)-0.5,
                         'xaxis': 'x2',
                         'mode': 'lines', 'hoverinfo': 'skip', 'showlegend': False,
                         'line': {'dash': 'dot', 'opacity': 0.0, 'color': 'white'}})
        return {'data': data_fig,
                'layout': {'title': {'text': 'Source visible during the observation',
                                     'y': 1, 'yanchor': 'top'},
                           'hovermode': 'closest',
                           'xaxis': {'title': 'Time (GST)', 'showgrid': False,
                                     'range': [gstimes[0], gstimes[-1]],
                                     # 'tickvals': np.arange(np.ceil(obs.gstimes.value[0]),
                                     #        np.floor(np.unwrap(obs.gstimes.value*2*np.pi/24)[-1]*24/(2*np.pi))+1),
                                     # 'ticktext': np.arange(np.ceil(obs.gstimes.value[0]),
                                     #        np.floor(np.unwrap(obs.gstimes.value*2*np.pi/24)[-1]*24/(2*np.pi))+1)%24,
                                     'tickformat': '%H:%M',
                                     'ticks': 'inside', 'showline': True, 'mirror': False,
                                     'hovermode': 'closest', 'color': 'black'},
                           'xaxis2': {'title': {'text': 'Time (UTC)', 'standoff': 0},
                                      'showgrid': False, 'overlaying': 'x', #'dtick': 1.0,
                                      'ticks': 'inside', 'showline': True, 'mirror': False,
                                      'hovermode': 'closest', 'color': 'black', 'side': 'top'},
                           'yaxis': {'ticks': '', 'showline': True, 'mirror': True,
                                     'showticklabels': False, 'zeroline': False,
                                     'showgrid': False, 'hovermode': 'closest',
                                     'startline': False}}}



    def get_fig_uvplane(self):
        data_fig = []
        bl_uv = self.get_uv_baseline()
        for bl_name in bl_uv:
            # accounting for complex conjugate
            uv = np.empty((2*len(bl_uv[bl_name]), 2))
            uv[:len(bl_uv[bl_name]), :] = bl_uv[bl_name]
            uv[len(bl_uv[bl_name]):, :] = -bl_uv[bl_name]
            data_fig.append({'x': uv[:,0],
                             'y': uv[:,1],
                             # 'type': 'scatter', 'mode': 'lines',
                             'type': 'scatter', 'mode': 'markers',
                             'marker': {'symbol': '.', 'size': 2},
                             'name': bl_name, 'hovertext': bl_name, 'hoverinfo': 'name', 'hovertemplate': ''})
        return {'data': data_fig,
                'layout': {'title': '', 'showlegend': False,
                           'hovermode': 'closest',
                           'width': 700, 'height': 700,
                           'xaxis': {'title': 'u (lambda)', 'showgrid': False, 'zeroline': False,
                                     'ticks': 'inside', 'showline': True, 'mirror': "all",
                                     'color': 'black'},
                           'yaxis': {'title': 'v (lambda)', 'showgrid': False, 'scaleanchor': 'x',
                                     'ticks': 'inside', 'showline': True, 'mirror': "all",
                                     'color': 'black', 'zeroline': False}}}



    def get_fig_dirty_map(self):
        # Right now I only leave the natural weighting map (the uniform does not always correspond to the true one)
        dirty_map_nat, laxis = self.get_dirtymap(pixsize=1024, robust='natural', oversampling=4)
        raise NotImplementedError
        # TODO: uncomment these two lines and move them outside observation. Flexibility
        # fig1 = px.imshow(img=dirty_map_nat, x=laxis, y=laxis[::-1], labels={'x': 'RA (mas)', 'y': 'Dec (mas)'}, \
        #         aspect='equal')
        # fig = make_subplots(rows=1, cols=1, subplot_titles=('Natural weighting',), shared_xaxes=True, shared_yaxes=True)
        fig.add_trace(fig1.data[0], row=1, col=1)
        mapsize = 30*self.synthesized_beam()['bmaj'].to(u.mas).value
        fig.update_layout(coloraxis={'showscale': False, 'colorscale': 'Inferno'}, showlegend=False,
                          xaxis={'autorange': False, 'range': [mapsize, -mapsize]},
                          yaxis={'autorange': False, 'range': [-mapsize, mapsize]}, autosize=False)
        fig.update_xaxes(title_text="RA (mas)", constrain="domain")
        fig.update_yaxes(title_text="Dec (mas)", scaleanchor="x", scaleratio=1)
        # dirty_map_nat, laxis = obs.get_dirtymap(pixsize=1024, robust='natural', oversampling=4)
        # dirty_map_uni, laxis = obs.get_dirtymap(pixsize=1024, robust='uniform', oversampling=4)
        # fig1 = px.imshow(img=dirty_map_nat, x=laxis, y=laxis[::-1], labels={'x': 'RA (mas)', 'y': 'Dec (mas)'}, \
        #         aspect='equal')
        # fig2 = px.imshow(img=dirty_map_uni, x=laxis, y=laxis[::-1], labels={'x': 'RA (mas)', 'y': 'Dec (mas)'}, \
        #         aspect='equal')
        # fig = make_subplots(rows=1, cols=2, subplot_titles=('Natural weighting', 'Uniform weighting'),
        #                     shared_xaxes=True, shared_yaxes=True)
        # fig.add_trace(fig1.data[0], row=1, col=1)
        # fig.add_trace(fig2.data[0], row=1, col=2)
        # mapsize = 30*obs.synthesized_beam()['bmaj'].to(u.mas).value
        # fig.update_layout(coloraxis={'showscale': False, 'colorscale': 'Inferno'}, showlegend=False,
        #                   xaxis={'autorange': False, 'range': [mapsize, -mapsize]},
        #                   # This xaxis2 represents the xaxis for fig2.
        #                   xaxis2={'autorange': False, 'range': [mapsize, -mapsize]},
        #                   yaxis={'autorange': False, 'range': [-mapsize, mapsize]}, autosize=False)
        # fig.update_xaxes(title_text="RA (mas)", constrain="domain")
        # fig.update_yaxes(title_text="Dec (mas)", row=1, col=1, scaleanchor="x", scaleratio=1)

        return fig

    # def export_to_pdf(self, outputname):
    #     """Exports the basic information of the observation into a PDF file.
    #     """
    # https://towardsdatascience.com/creating-pdf-files-with-python-ad3ccadfae0f
    #     from fpdf import FPDF
    #     pdf = FPDF(orientation='P', unit='cm', format='A4')
    #     # pdf_w = 21.0
    #     # pdf_h = 29.7
    #     pdf.add_page()
    #
    #     plotly.io.write_image(pltx,file='pltx.png',format='png',width=700, height=450)
    #     pltx=(os.getcwd()+'/'+"pltx.png")
    #
    #     pdf.image(sctplt,  link='', type='', w=1586/80, h=1920/80)
    #     self.set_font('Arial', 'B', 16)
    #     self.set_text_color(220, 50, 50)
    #     self.cell(w=210.0, h=40.0, align='C', txt="LORD OF THE PDFS", border=0)
    #     self.set_font('Arial', '', 12)
    #     self.multi_cell(0,10,txt)
    #
    #     pdf.output('test.pdf','F')
    #





import asyncio
from typing import Optional, Union, Iterable, Sequence, Self
from importlib import resources
from functools import partial
from itertools import product
import subprocess
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
import numpy as np
import yaml
from rich import print as rprint
# import scipy.ndimage
import datetime as dt
from enum import Enum, auto
from dataclasses import dataclass
from astropy import units as u
from astropy import coordinates as coord
from astropy.time import Time, TimeDelta
from ortools.sat.python import cp_model
from .stations import Station, Stations
from .sources import Source, Scan, ScanBlock, SourceType, SourceNotVisible
from .observation import Observation


def _t2mjd10(time: Time) -> int:
    """Converts an astropy.time.Time object into a decimal MJD integer value, meaning int(MJD*10).
    """
    return int(time.mjd * 10)


def _t2time(time: int) -> Time:
    """Converts back a int(MJD*10) time stampt into a astropy.time.Time object.
    """
    return Time(time/10.0, format='mjd').datetime


class ScanBlockScheduler:
    def __init__(self, observation: Observation):
        """
        Args:
            observation (Observation): The observation to be schedued.
        """
        self.observation = observation
        self.starttime = _t2mjd10(observation.times[0])
        self.endtime = _t2mjd10(observation.times[-1])
        self.max_duration = self.endtime - self.starttime
        self.model = cp_model.CpModel()
        self.solver = cp_model.CpSolver()
        self.solution = None

    def create_variables(self):
        self.start_times = {}
        self.durations = {}
        self.end_times = {}

        for i, block in self.observation.scans.items():
            self.start_times[i] = self.model.NewIntVar(self.starttime, self.endtime,
                                                       f'start_time_{i}')
            self.durations[i] = self.model.NewIntVar(int(10*np.sum([s.duration.to(u.h).value for s in block.scans])),
                                                     self.max_duration, f'duration_{i}')
            self.end_times[i] = self.model.NewIntVar(self.starttime, self.endtime,
                                                     f'end_time_{i}')

            # Link start, duration, and end times
            self.model.Add(self.end_times[i] == self.start_times[i] + self.durations[i])

    def add_constraints(self):
        # Ensure all scan blocks are within observation window
        for i in self.observation.scans.keys():
            self.model.Add(self.start_times[i] >= self.starttime)
            self.model.Add(self.end_times[i] <= self.endtime)

        # Non-overlapping constraint
        for ii, i in enumerate(self.observation.scans.keys()):
            for j in list(self.observation.scans.keys())[ii+1:]:
                self.model.AddNoOverlap([
                    self.model.NewIntervalVar(self.start_times[i], self.durations[i], self.end_times[i],
                                              f'interval_{i}'),
                    self.model.NewIntervalVar(self.start_times[j], self.durations[j], self.end_times[j],
                                              f'interval_{j}')
                ])

    def optimize_station_participation(self):
        # BUG: I don't think this works as it is!
        for i, block in self.observation.scans.items():
            participation_var = self.model.NewIntVar(0, len(self.stations), f'participation_{i}')
            self.model.Add(participation_var == sum(
                           self.observation.is_observable_at(self.self.start_times[i]).values()
                           ))
            self.model.Maximize(participation_var)

    def optimize_coordinate_separation(self):
        # BUG: this also doesn't work.  How I compute one block and the next one in this context?
        for i, block in self.observation.scans.items():
            separation = self.model.NewIntVar(0, 360, f'separation_{i}')
            self.model.Add(separation == int(abs(
                self.observation.scans[i].sources(SourceType.TARGET)[0].coordinates -
                self.observation.scans[i+1].sources(SourceType.TARGET)[0].coordinates
            )))
            self.model.Minimize(separation)

    def optimize_elevation(self):
        time_index = lambda i: np.where(self.observation.times.mjd == min(self.observation.times.mjd,
                                        key=lambda x: abs(x-self.start_times[i]/10)))
        for i, elevations in enumerate(self.observation.elevations()):
            for station, elev in self.observation.elevations()[elevations].items():
                elevation = self.model.NewIntVar(0, 90, f'elevation_{i}_{station}')
                self.model.Add(elevation == elev[time_index(i)])
                self.model.Add(elevation >= 20)
                self.model.Add(elevation <= 82)

    def solve(self):
        self.create_variables()
        self.add_constraints()
        # self.optimize_station_participation()
        # self.optimize_elevation()
        # self.optimize_coordinate_separation()

        status = self.solver.Solve(self.model)

        if status == cp_model.OPTIMAL or status == cp_model.FEASIBLE:
            schedule = []
            for i, block in self.observation.scans.items():
                start = self.solver.Value(self.start_times[i])
                duration = self.solver.Value(self.durations[i])
                schedule.append((i, start, duration))

            self.solution = schedule
        else:
            self.solution = None

    def print_schedule(self):
        for block, start, duration in self.solution:
            print(f"ScanBlock {block}: Start={_t2time(start)}, Duration={duration}")


# scheduler = ScanBlockScheduler(scan_blocks, stations, observation_start, observation_end)
# optimal_schedule = scheduler.solve()
#
# if optimal_schedule:
#     for block, start, duration in optimal_schedule:
#         print(f"ScanBlock {block.id}: Start={start}, Duration={duration}")
# else:
#     print("No feasible schedule found.")



import pytest
import configparser
import numpy as np
from importlib import resources
from astropy import coordinates as coord
from astropy import units as u
from astropy.time import Time
from astroplan import FixedTarget
from vlbiplanobs import stations
from vlbiplanobs import freqsetups


def test_stations_catalog():
    """Checks that the stations_catalog.inp file is self-consistent.
    So no duplicated stations (with the same stationname, as that must be unique), or wrong parameters.
    """
    config = configparser.ConfigParser()
    with resources.as_file(resources.files("vlbiplanobs.data").joinpath("stations_catalog.inp")) as stat_cat:
        config.read(stat_cat)

    codenames = set()
    for stationname in config.sections():
        for a_key in ('station', 'code', 'networks', 'country', 'diameter',
                      'real_time', 'position'):
            assert a_key in config[stationname], \
                   f"'{a_key}' is not present in the stations_catalog.inp for {stationname}"

        assert config[stationname]['code'] not in codenames, \
               f"Duplicated codename {config[stationname]['code']} in stations_catalog.inp."
        codenames.add(config[stationname]['code'])
        temp = [float(i.strip()) for i in config[stationname]['position'].split(',')]
        assert len(temp) == 3, f"The station position of {stationname} does not have the three expected numbers."
        _ = coord.EarthLocation(u.Quantity(float(temp[0]), u.m), u.Quantity(float(temp[1]), u.m),
                                u.Quantity(float(temp[2]), u.m))
        # Getting the SEFD values for the bands
        assert config[stationname]['real_time'] in ('yes', 'no'), \
               f"'real_time' must be 'yes' or 'no' for {stationname}"
        # At least one SEFD per station
        has_sefd = False
        for akey in config[stationname].keys():
            has_sefd = ('SEFD_' in akey.upper()) or has_sefd

        assert has_sefd, f"Station {stationname} must have at least a SEFD value for one frequency."


def test_network_catalog():
    """Checks the consistency of the information in the  network_catalog.inp file.
    """
    config = configparser.ConfigParser()
    with resources.as_file(resources.files("vlbiplanobs.data").joinpath("network_catalog.inp")) as net_cat_path:
        config.read(net_cat_path)

    # To verify that all stations (codenames) defined in the network file as defined in the stations_catalog
    all_stations = stations.Stations().station_codenames
    for networkname in config.sections():
        for a_key in ('name', 'default_antennas', 'max_datarate', 'observing_bands'):
            assert a_key in config[networkname], \
                   f"'{a_key}' is not present in the network_catalog.inp for {networkname}"
        network_antennas = [a.strip() for a in config[networkname]['default_antennas'].split(',')]
        for antenna in network_antennas:
            assert antenna in all_stations, \
                   f"{antenna}, defined in {networkname} is not present in stations_catalog.inp"


def test_station_init():
    """Tests the Station class and different possibilities during the creation of the station.
    """
    sefds = {'18cm': 100, '6cm': 40, '0.1cm': 200}
    _ = stations.Station('name', 'Nm', 'VLBI', coord.EarthLocation(0., 0., 0.), sefds)
    _ = stations.Station('name', 'Nm', ('EVN', 'VLBA'), coord.EarthLocation(0., 0., 0.), {})
    with pytest.raises(AssertionError):
        # Wrong names
        _ = stations.Station(None, 'Nm', 'VLBI', coord.EarthLocation(0., 0., 0.), {}, 10)
        _ = stations.Station(5, 'Nm', 'VLBI', coord.EarthLocation(0., 0., 0.), {}, 10)
        _ = stations.Station('name', None, 'VLBI', coord.EarthLocation(0., 0., 0.), {}, 10)
        _ = stations.Station('name', True, 'VLBI', coord.EarthLocation(0., 0., 0.), {}, 10)
        _ = stations.Station('name', 'Nm', None, coord.EarthLocation(0., 0., 0.), {}, 10)
        _ = stations.Station('name', 'Nm', 5.0, coord.EarthLocation(0., 0., 0.), {}, 10)
        # Negative min_elevations
        _ = stations.Station('name', 'Nm', 'VLBI', coord.EarthLocation(0., 0., 0.), {}, -10)
        # Wrong units min_elevations
        _ = stations.Station('name', 'Nm', 'VLBI', coord.EarthLocation(0., 0., 0.), {}, 10*u.m)


# TODO: add tests on: check Mount type is what is written in the file, and depending on that, if it got the
# correct constraints (e.g. HourAngle for EQUAT, etc)

def test_station_functions():
    """Tests the Station functions.
    """
    sefds = {'18cm': 100*u.Jy, '6cm': 40*u.Jy, '0.1cm': 200*u.Jy}
    a_station = stations.Station('name', 'Nm', 'VLBI',
                                 coord.EarthLocation(3839348.973*u.m, 430403.51*u.m, 5057990.099*u.m), sefds)
    assert isinstance(a_station.name, str)
    assert isinstance(a_station.fullname, str)
    assert a_station.fullname == a_station.name
    assert a_station.networks == ('VLBI',)
    assert len(a_station.networks) == 1
    assert isinstance(a_station.country, str)
    assert isinstance(a_station.diameter, str)
    assert isinstance(a_station.real_time, bool)
    assert isinstance(a_station.location, coord.EarthLocation)
    assert a_station.location == coord.EarthLocation(3839348.973*u.m, 430403.51*u.m, 5057990.099*u.m)
    assert list(a_station.bands) == ['18cm', '6cm', '0.1cm']
    assert isinstance(a_station.sefds, dict)
    assert a_station.sefds == sefds
    assert a_station.has_band('18cm')
    assert not a_station.has_band('45cm')
    assert a_station.sefd('18cm') == 100*u.Jy
    with pytest.raises(KeyError):
        a_station.sefd('45')

    times1 = Time('2020-03-21 1:00') + np.arange(0, 2*60, 10)*u.min
    times2 = Time('2020-03-21 3:00') + np.arange(0, 4*60, 10)*u.min
    src1 = FixedTarget(coord=coord.SkyCoord('0h0m0s 30d0m0s'), name='testSrc')
    # a_station.elevation(times1, src1)  # Should have elevation ranging -7 to 1.1 deg.
    # a_station.elevation(times2, src1)  # Should have elevation ranging 2.2 to 34 deg.
    assert not all(a_station.is_observable(times1, src1))
    assert all(a_station.is_observable(times2, src1))
    assert not a_station.is_always_observable(times2, src1)
    assert len(a_station.elevation(times2, src1)) == len(times2)
    assert len(a_station.elevation(times1, src1)) == len(times1)
    assert np.equal(a_station.elevation(times2, src1).value, a_station.altaz(times2, src1).alt.value)[0]


def test_station_file():
    """Tests the consistency between the antennas in the station catalog, network catalog, and freq setups.
    Mainly:  no duplicated antenna code names, all antennas in the networks need to be defined, and all
    setups in the antennas need to be defined in the freqsetup.py
    """
    # all_stations = list(stations.Stations.get_stations_from_configfile())
    all_stations = stations.Stations()
    assert len(all_stations) > 0

    # Verifying networks - it would raise issues if inconsistenties
    all_networks = stations.Stations.get_networks_from_configfile()

    # double checking with one Station in particular
    with open(resources.files("vlbiplanobs.data").joinpath("stations_catalog.inp")) as afile:
        in_ef = False
        n_ants = 0
        for aline in afile.readlines():
            if 'station=' in aline.strip().replace(' ', ''):
                n_ants += 1

            if in_ef and ('station=' in aline.strip().replace(' ', '')):
                # We finished with effelsberg!
                in_ef = False

            if in_ef and (len(aline.strip()) > 0) and (aline.strip()[0] != '#') and ('[' not in aline):
                key, value = [a.strip() for a in aline.split('=')]
                match key:
                    case 'station':
                        assert value == all_stations['Ef'].fullname
                    case 'network':
                        assert value == all_stations['Ef'].network
                    case 'possible_networks':
                        for avalue in value.split(','):
                            assert avalue.strip() in all_stations['Ef'].all_networks
                    case 'country':
                        assert value == all_stations['Ef'].country
                    case 'diameter':
                        assert value == all_stations['Ef'].diameter
                    case 'position':
                        avalue = [float(a.strip()) for a in value.split(',')]
                        assert avalue[0]*u.m == all_stations['Ef'].location.x
                        assert avalue[1]*u.m == all_stations['Ef'].location.y
                        assert avalue[2]*u.m == all_stations['Ef'].location.z
                    case 'mount':
                        assert value == all_stations['Ef'].mount.mount_type.name
                    case 'ax1rate':
                        assert float(value)*u.deg/u.s == all_stations['Ef'].mount.ax1.speed
                    case 'ax2lim':
                        avalue = [float(a.strip())*u.deg for a in value.split(',')]
                        assert avalue[0] == all_stations['Ef'].mount.ax2.limits[0]
                        assert avalue[1] == all_stations['Ef'].mount.ax2.limits[1]
                    case 'ax1acc':
                        assert float(value)*u.deg/u.s/u.s == all_stations['Ef'].mount.ax1.acceleration
                    case 'real_time':
                        assert bool(value) == all_stations['Ef'].real_time

                if 'SEFD' in key:
                    assert all_stations['Ef'].has_band(key.removeprefix('SEFD_') + 'cm')
                    assert all_stations['Ef'].sefd(key.removeprefix('SEFD_')) == float(value)
                    assert key.removeprefix('SEFD_') + 'cm' in freqsetups.bands

            if aline.strip() == 'station = Effelsberg':
                in_ef = True

            # ingore if we are at any other station, but:
            # At the same time, I check for consistency in the network names
            if aline.strip().replace(' ', '') == 'network=':
                value = aline.split('=')[1].strip()
                assert ',' not in value
                assert value in all_networks, "The network '{value}' is not defined in the networks file."

            if aline.strip().replace(' ', '') == 'possible_network=':
                values = [a.strip() for a in aline.split('=')[1].strip().split(',')]
                for avalue in values:
                    assert avalue in all_networks, "The network '{avalue}' is not defined in the networks file."

            if aline.strip().replace(' ', '') == 'maxdatarate_':
                assert aline.split('=')[0].strip().removeprefix('maxdatarate_') in all_networks

    # This value may change in the future but then is a problem of the test, not the code
    assert 'EVN' in all_stations['De'].max_datarate, \
           f"De doesn't have max_datarate for EVN (keys: {all_stations['De'].max_datarate})"
    assert all_stations['De'].max_datarate['EVN'] == 512*u.Mb/u.s, \
           f"De has a mismatching data rate of {all_stations['De'].max_datarate['EVN']}."
    assert n_ants == len(all_stations)

    # Check that all stations have an associated image
    for ant in all_stations:
        with resources.as_file(resources.files("vlbiplanobs.gui.assets").
                               joinpath(f"ant-{ant.name.replace(' ', '_').lower()}.jpg")) as antfile:
            assert antfile.exists()

    for net in all_networks:
        with resources.as_file(resources.files("vlbiplanobs.gui.assets").
                               joinpath(f"network-{net.replace(' ', '_').lower()}.png")) as netfile:
            if net != 'e-EVN':
                assert netfile.exists()



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class Checkbox(Component):
    """A Checkbox component.
Creates a single checkbox input. Use the `checked` prop in your callbacks.

Keyword arguments:
- id (string; optional): The ID of this component, used to identify dash components
in callbacks. The ID needs to be unique across all of the
components in an app.
- checked (boolean; optional): Whether Checkbox has been checked or not
- className (string; optional): The class of the container (div)
- style (dict; optional): The style of the container (div)
- key (string; optional): A unique identifier for the component, used to improve
performance by React.js while rendering components
See https://reactjs.org/docs/lists-and-keys.html for more info
- loading_state (dict; optional): Object that holds the loading state object coming from dash-renderer.
  loading_state has the following type: dict containing keys 'is_loading', 'prop_name', 'component_name'.
Those keys have the following types:
  - is_loading (boolean; optional): Determines if the component is loading or not
  - prop_name (string; optional): Holds which property is loading
  - component_name (string; optional): Holds the name of the component that is loading
- persistence (boolean | string | number; optional):
  Used to allow user interactions in this component to be persisted when
the component - or the page - is refreshed. If `persisted` is truthy and
hasn't changed from its previous value, a `value` that the user has
changed while using the app will keep that change, as long as
the new `value` also matches what was given originally.
Used in conjunction with `persistence_type`.
- persisted_props (list of a value equal to: 'checked's; default ['checked']):
  Properties whose user interactions will persist after refreshing the
component or the page. Since only `value` is allowed this prop can
normally be ignored.
- persistence_type (a value equal to: 'local', 'session', 'memory'; default 'local'):
  Where persisted user changes will be stored:
memory: only kept in memory, reset on page refresh.
local: window.localStorage, data is kept after the browser quit.
session: window.sessionStorage, data is cleared once the browser quit."""
    @_explicitize_args
    def __init__(self, id=Component.UNDEFINED, checked=Component.UNDEFINED, disabled=Component.UNDEFINED,
                 className=Component.UNDEFINED, style=Component.UNDEFINED, key=Component.UNDEFINED,
                 loading_state=Component.UNDEFINED, persistence=Component.UNDEFINED,
                 persisted_props=Component.UNDEFINED, persistence_type=Component.UNDEFINED, **kwargs):
        self._prop_names = ['id', 'checked', 'disabled', 'className', 'style', 'key', 'loading_state', 'persistence', 'persisted_props', 'persistence_type']
        self._type = 'Checkbox'
        self._namespace = 'dash_bootstrap_components'
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'checked', 'disabled', 'className', 'style', 'key',
                                     'loading_state', 'persistence', 'persisted_props', 'persistence_type']
        self.available_wildcard_properties = []

        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs
        args = {k: _locals[k] for k in _explicit_args if k != 'children'}

        for k in []:
            if k not in args:
                raise TypeError(
                    'Required argument `' + k + '` was not specified.')
        super(Checkbox, self).__init__(**args)



from typing import Optional
from astropy import units as u
from astropy.time import Time
from astroplan import Observer, FixedTarget
from astroplan import Constraint, AltitudeConstraint, SunSeparationConstraint, \
                      MoonSeparationConstraint, min_best_rescale

__all__: list[str] = ['ElevationConstraint', 'SunSeparationConstraint', 'MoonSeparationConstraint',
                      'AzimuthConstraint', 'HourAngleConstraint', 'DeclinationConstraint']

# Limits for the station to check if it can observe, adapted for VLBI
# - SunSeparationConstraint(min, max=None)
# - AltitudeConstraint(min, max)
# - MoonSeparationConstraint(min, max=None)


class ElevationConstraint(AltitudeConstraint):
    """Constraint on the elevation of the source.
    """
    pass


class AzimuthConstraint(Constraint):
    """Constraint the azimuth for the station
    """

    def __init__(self, min: Optional[u.Quantity] = None, max: Optional[u.Quantity] = None,
                 boolean_constraint: bool = True):
        """
        min : `~astropy.units.Quantity` or `None` (optional)
            Minimum acceptable azimuth for the target. `None` indicates no limit.
        max : `~astropy.units.Quantity` or `None` (optional)
            Minimum acceptable azimuth for the target. `None` indicates no limit.
        """
        self.min: u.Quantity = min if min is not None else -10*u.deg
        self.max: u.Quantity = max if max is not None else 370*u.deg
        self.boolean_constraint: bool = boolean_constraint

    def compute_constraint(self, times: Time, observer: Observer, targets: FixedTarget):
        azimuths = observer.altaz(times, targets).az
        if self.boolean_constraint:
            return ((self.min < azimuths) & (azimuths < self.max))
        else:
            rescale = min_best_rescale(azimuths, self.min, self.max, less_than_min=0)
            return rescale


class HourAngleConstraint(Constraint):
    """Constraint the hour angle for the station.
    """

    def __init__(self, min: Optional[u.Quantity] = None, max: Optional[u.Quantity] = None,
                 boolean_constraint: bool = True):
        """
        min : `~astropy.units.Quantity` or `None` (optional)
            Minimum acceptable hour angle for the target. `None` indicates no limit.
        max : `~astropy.units.Quantity` or `None` (optional)
            Maximum acceptable hour angle for the target. `None` indicates no limit.
        """
        self.min = min if min is not None else -370*u.deg
        self.max = max if max is not None else 370*u.deg
        self.boolean_constraint = boolean_constraint

    def compute_constraint(self, times: Time, observer: Observer, targets: FixedTarget):
        hour_angles = observer.target_hour_angle(times, targets)
        if self.boolean_constraint:
            return ((self.min < hour_angles) & (hour_angles < self.max)) | \
                   (((self.min + 24*u.hourangle) < hour_angles) & (hour_angles < (self.max + 24.0*u.hourangle)))
        else:
            rescale = min_best_rescale(hour_angles, self.min, self.max, less_than_min=0)
            return rescale


class DeclinationConstraint(Constraint):
    """Constraint the source declination for the station.
    """

    def __init__(self, min: Optional[u.Quantity] = None, max: Optional[u.Quantity] = None,
                 boolean_constraint: bool = True):
        """
        min : `~astropy.units.Quantity` or `None` (optional)
            Minimum acceptable declination for the target. `None` indicates no limit.
        max : `~astropy.units.Quantity` or `None` (optional)
            Maximum acceptable declination for the target. `None` indicates no limit.
        """
        self.min = min if min is not None else -90*u.deg
        self.max = max if max is not None else 91*u.deg
        self.boolean_constraint = boolean_constraint

    def compute_constraint(self, times: Time, observer: Observer, targets: FixedTarget):
        target_declination = targets.dec
        if self.boolean_constraint:
            return ((self.min < target_declination) & (target_declination < self.max))
        else:
            rescale = min_best_rescale(target_declination, self.min, self.max, less_than_min=0)
            return rescale
